import time, asyncio, ujson as json, redis
from seneca.libs.logger import get_logger
from seneca.engine.interface import SenecaInterface
from seneca.constants.config import *
from seneca.engine.conflict_resolution import CRContext
from seneca.engine.book_keeper import BookKeeper
from seneca.engine.util import module_path_for_contract
from collections import deque, defaultdict
from typing import Callable, List


SUCC_FLAG = 'SUCC'


class Macros:
    # TODO we need to make sure these keys dont conflict with user stuff in the common layer. I.e. users cannot be
    # creating keys named '_execution' or '_conflict_resolution'
    EXECUTION = '_execution'
    CONFLICT_RESOLUTION = '_conflict_resolution'

    ALL_MACROS = [EXECUTION, CONFLICT_RESOLUTION]


class Phase:
    EXEC_TIMEOUT = 14  # Number of seconds client will wait for other clients to finish execution phase
    CR_TIMEOUT = 14  # Number of seconds client will wait for other clients to finish conflict resolution phase
    BLOCK_TIMEOUT = 30  # Number of seconds a pending db will wait until it is at the top (first element) of pending_dbs
    POLL_INTERVAL = 0.5  # Poll for Phase changes every POLL_INTERVAL seconds

    @staticmethod
    def reset_phase_variables(db):
        db.set(Macros.EXECUTION, 0)
        db.set(Macros.CONFLICT_RESOLUTION, 0)

    @staticmethod
    def incr_phase_variable(db, key):
        db.incr(key)

    @staticmethod
    def get_phase_variable(db, key):
        if not db.exists(key):
            return 0
        return int(db.get(key).decode())


class SenecaClient(SenecaInterface):

    def __init__(self, sbb_idx, num_sbb, concurrent_mode=True, loop=None):
        # TODO do we even need to bother with the concurrent_mode flag? We are treating that its always true --davis
        super().__init__()

        name = self.__class__.__name__ + "[sbb-{}]".format(sbb_idx)
        self.log = get_logger(name)
        self.loop = loop or asyncio.get_event_loop()

        self.port = get_redis_port()
        self.password = get_redis_password()

        self.sbb_idx = sbb_idx
        self.num_sb_builders = num_sbb
        self.concurrent_mode = concurrent_mode
        self.max_number_workers = NUM_CACHES

        self.master_db = None  # A redis.StrictRedis instance
        self.active_db = None  # Set to a CRContext instance
        self.available_dbs = deque()  # List of CRContext instances
        self.pending_dbs = deque()  # List of CRContext instances
        self.pending_futures = {}  # Map of input hashes -> {'fut': asyncio.Future, 'data': CRContext}

        self._setup_dbs()

    def _setup_dbs(self):
        self.master_db = redis.StrictRedis(host='localhost', port=self.port, db=MASTER_DB, password=self.password)
        for db_num in range(self.max_number_workers):
            db_client = redis.StrictRedis(host='localhost', port=self.port, db=db_num+DB_OFFSET, password=self.password)
            cr_data = CRContext(working_db=db_client, master_db=self.master_db, sbb_idx=self.sbb_idx)
            self.available_dbs.append(cr_data)

    def _reset_cr_data(self, ds: CRContext):
        ds.reset()
        Phase.reset_phase_variables(ds.working_db)

    def flush_all(self):
        """ Flushes all pending/active/available dbs. This effectively 'resets' all databases except master."""
        if self.active_db:
            self._reset_cr_data(self.active_db)
            self.active_db = None
        for s in (self.pending_dbs, self.available_dbs):
            for cr in s:
                self._reset_cr_data(cr)
            s.clear()

        for input_hash in self.pending_futures:
            self.pending_futures[input_hash]['fut'].cancel()
        self.pending_futures.clear()

        self._setup_dbs()

    def _update_master_db(self):
        assert len(self.pending_dbs) > 0, "No pending dbs to update to master!"

        cr_data = self.pending_dbs.popleft()
        assert cr_data.merged_to_common, "CRData not merged to common yet!"

        if self.sbb_idx == 0:
            assert Phase.get_phase_variable(cr_data.working_db, Macros.EXECUTION) == self.num_sb_builders, \
                "Execution stage incomplete!"
            assert Phase.get_phase_variable(cr_data.working_db, Macros.CONFLICT_RESOLUTION) == self.num_sb_builders, \
                "Conflict resolution stage incomplete!"

            self.log.notice("Merging common layer to master db")
            CRContext.merge_to_master(working_db=cr_data.working_db, master_db=self.master_db)

            self.log.debugv("Resetting CRContext with input hash {}".format(cr_data.input_hash))
            self._reset_cr_data(cr_data)

        self.available_dbs.append(cr_data)

    def update_master_db(self, input_hash: str):
        assert len(self.pending_dbs) > 0, "No pending dbs to update to master!"
        assert input_hash in self.pending_futures, "Input hash {} not in pending_futures {}".format(input_hash, self.pending_futures)
        cr_data = self.pending_futures[input_hash]['data']
        assert cr_data in self.pending_dbs, "you done shit the bed again davis"

        cr_finished = Phase.get_phase_variable(cr_data.working_db, Macros.CONFLICT_RESOLUTION) == self.num_sb_builders
        if not cr_data.merged_to_common or (self.sbb_idx == 0 and cr_finished):
            self.log.notice(("Deferring merge for input_hash {}".format(input_hash)))
            self.pending_futures[cr_data.input_hash]['merge'] = True
        else:
            self._update_master_db()

    def submit_contract(self, contract):
        self.publish_code_str(contract.contract_name, contract.sender, contract.code, scope={
            'rt': {
                'author': contract.sender,
                'sender': contract.sender,
                'contract': contract.contract_name
            }
        })

    def run_contract(self, contract):
        assert self.active_db, "active_db must be set to run a contract. Did you call _start_sb?"
        assert self.active_db.input_hash, "Input hash not set...davis u done goofed again"

        result = self._run_contract(contract, contract_idx=self.active_db.next_contract_idx, data=self.active_db)
        self.active_db.add_contract_result(contract, result)

    def _run_contract(self, contract, contract_idx: int, data: CRContext) -> str:
        """ Runs the contract object, and retuns a string representing the result (succ/fail).
        Note: Assumes the BookKeeping info has already been set. """
        BookKeeper.set_info(sbb_idx=self.sbb_idx, contract_idx=contract_idx, data=data)
        contract_name = contract.contract_name

        try:
            # Super sketch hack to differentiate between ContractTransactions and PublishTransactions
            if hasattr(contract, 'contract_code'):
                author = contract.sender
                self.publish_code_str(fullname=contract.contract_name, author=author,
                                      code_str=contract.contract_code)
            else:
                author = self.get_contract_meta(contract_name)['author']
                mod_path = module_path_for_contract(contract)
                self.execute_function(module_path=mod_path, sender=contract.sender,
                                      stamps=contract.stamps, **contract.kwargs)
            result = SUCC_FLAG

        except Exception as e:
            self.log.warning("Contract failed with error: {} \ncontract obj: {}".format(e, contract))
            result = 'FAIL' + ' -- ' + str(e)
            data.rollback_contract(contract_idx)

        return result

    def _rerun_contracts_for_cr_data(self, cr_data: CRContext):
        """ Reruns any contracts in cr_data, if necessary. This should be done before we merge cr_data to common. """
        self.log.info("Rerunning any necessary contracts for CRData with input hash {}".format(cr_data.input_hash))
        for i in cr_data.iter_rerun_indexes():
            result = self._run_contract(cr_data.contracts[i], contract_idx=i, data=cr_data)
            cr_data.update_contract_result(contract_idx=i, result=result)

    def catchup(self):
        raise NotImplementedError('code this up if ur tryna use it u lazy bum')

    def has_available_db(self) -> bool:
        return len(self.available_dbs) > 0

    def execute_sb(self, input_hash: str, contracts: list, completion_handler: Callable[[CRContext], None]):
        # TODO -- wait until we have an available DB. If we do not have any available db's put this call in a queue
        # and pull it off the queue once we pop a pending_db and gain another available_db

        self._start_sb(input_hash)
        for c in contracts:
            self.run_contract(c)
        self._end_sb(completion_handler)

    def _start_sb(self, input_hash: str):
        assert self.active_db is None, "Attempted to _start_sb, but active_db is already set! Did you end the " \
                                       "previous subblock with _end_sb?"
        if not self.has_available_db():
            raise Exception("Attempted to start a new sub block, but there are no available DBs!")

        self.active_db = self.available_dbs.popleft()
        self.active_db.input_hash = input_hash

    def _end_sb(self, completion_handler: Callable[[CRContext], None]):
        """
        Ends the current sub block. Specifically, this method:
         - Increments the EXECUTION phase variable
         - Pushes active_db onto the top of the pending_db stack
         - Ensures the _wait_and_merge_to_common future (see doc string on that func for more details)
         - Once the _wait_and_merge_to_common future is done, the callback is trigger
        """
        assert self.active_db, "Active db not set! Did you call _start_sb?"
        self.log.notice("Ending sub block {} which has input hash {}".format(self.sbb_idx, self.active_db.input_hash))

        Phase.incr_phase_variable(self.active_db.working_db, Macros.EXECUTION)

        future = asyncio.ensure_future(self._wait_and_merge_to_common(self.active_db, completion_handler))
        self.pending_futures[self.active_db.input_hash] = {'fut': future, 'data': self.active_db, 'merge': False}

        self.pending_dbs.append(self.active_db)
        self.active_db = None  # we really don't care, but might be useful initially for error checking

    async def _wait_and_merge_to_common(self, cr_data: CRContext, completion_handler: Callable[[CRContext], None]):
        """
        - Waits for all other SBBs to finish execution. Raises an error if this takes longer than Phase.EXEC_TIMEOUT
        - Waits for this cr_data to be first in line in 'pending_dbs' (bottom of stack)
        - Waits for all subblocks before it finish conflict resolution. Raises an error if it takes longer than Phase.CR_TIMEOUT
        - Starts conflict resolution, which involves:
            - Rerun any necessary contracts
            - Merges cr_data to common layer, and then increment the CONFLICT_RESOLUTION phase variable
        - If the 'merge' flag is set by an earlier update_master_db call, then this is done next
        """
        await self._wait_for_execution_stage(cr_data)

        if self.pending_dbs[0] is not cr_data:
            await self._wait_until_top_of_pending(cr_data)

        await self._wait_for_cr_and_merge(cr_data)

        self.log.notice("Invoking completion handler for sub block with input hash {}".format(cr_data.input_hash))
        completion_handler(cr_data)

        if self.pending_futures[cr_data.input_hash]['merge']:
            self.log.debugv("Merge flag set to true for sbb with input hash {}".format(cr_data.input_hash))
            # If this is the first sub-block, then we are responsible for merging to master. Thus we must wait
            # for everyone to finish conflict resolution before we merge to master.
            if self.sbb_idx == 0:
                await self._wait_for_everybody_cr(cr_data)
            self._update_master_db()

        # self.pending_futures[cr_data.input_hash]['fut'].set_result('done')  # TODO fix this its complaining...

    async def _wait_for_cr_and_merge(self, cr_data: CRContext):
        self.log.info("Waiting for other SBBs to finish conflict resolution...")
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.CONFLICT_RESOLUTION, value=self.sbb_idx,
                                            timeout=(len(self.pending_dbs) + 1) * Phase.CR_TIMEOUT)
        self.log.info("Done waiting for other SBBs to finish contract resolution")

        self._rerun_contracts_for_cr_data(cr_data)
        self.log.notice("Merging sbb_{} data to common layer".format(self.sbb_idx))
        cr_data.merge_to_common()
        Phase.incr_phase_variable(cr_data.working_db, Macros.CONFLICT_RESOLUTION)

        self.log.debug("Finished finalizing sub block for input hash {}!".format(cr_data.input_hash))

    async def _wait_until_top_of_pending(self, cr_data: CRContext):
        """ Waits until cr_data is at the top (first element) of self.pending_dbs """
        # TODO technically this 'wait' can be triggered reactively when we leftpop pending_dbs in merge_to_master
        # This is an optimization we can do later
        elapsed = 0
        timeout = Phase.BLOCK_TIMEOUT * len(self.pending_dbs)
        self.log.debug("CRData with input hash {} waiting its turn until its at the top of pending_dbs..."
                       .format(cr_data.input_hash))

        while self.pending_dbs[0] is not cr_data:
            await asyncio.sleep(Phase.POLL_INTERVAL)
            elapsed += Phase.POLL_INTERVAL

            if elapsed >= timeout:
                err_msg = "Exceeded timeout of {} waiting for db with input hash {} to reach top of pending dbs!" \
                          .format(timeout, cr_data.input_hash)
                self.log.fatal(err_msg)
                raise Exception(err_msg)

        self.log.info("CRData with input hash {} DONE waiting to be at the top of pending db!".format(cr_data.input_hash))

    async def _wait_for_execution_stage(self, cr_data: CRContext):
        self.log.info("Waiting for other SBBs to finish execution...")
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.EXECUTION, value=self.num_sb_builders,
                                            timeout=(len(self.pending_dbs) + 1) * Phase.EXEC_TIMEOUT)
        self.log.info("Done waiting for other SBBs to finish execution")

    async def _wait_for_everybody_cr(self, cr_data: CRContext):
        self.log.info("Waiting for conflict resolution to finish for ALL sub blocks...")
        await self._wait_for_phase_variable(db=cr_data.working_db, key=Macros.CONFLICT_RESOLUTION,
                                            value=self.num_sb_builders, timeout=Phase.CR_TIMEOUT)
        self.log.info("Conflict resolution complete for ALL sub blocks")

    async def _wait_for_phase_variable(self, db: redis.StrictRedis, key: str, value: int, timeout: int):
        elapsed = 0
        while Phase.get_phase_variable(db, key) != value:
            await asyncio.sleep(Phase.POLL_INTERVAL)
            elapsed += Phase.POLL_INTERVAL

            if elapsed >= timeout:
                err_msg = "Client with sbb_idx {} exceeded timeout of {} waiting for phase key {} to reach {}! " \
                          "Current conflict resolution phase value: {}" \
                    .format(self.sbb_idx, elapsed, key, value,
                            Phase.get_phase_variable(db, key))
                self.log.fatal(err_msg)
                raise Exception(err_msg)

        self.log.debug("Waited a total of {} seconds for phase variable {} to reach value {}".format(elapsed, key, value))



from functools import wraps
from inspect import signature, _empty
from collections import namedtuple
import ast


def module_path_for_contract(contract) -> str:
    return "seneca.contracts.{}.{}".format(contract.contract_name, contract.func_name)


def make_n_tup(d):
    # TODO: make sure this is good/safe
    return namedtuple('_', ' '.join(d.keys()))(**d)

def auto_set_fields(f):
    f_sig = signature(f)

    @wraps(f)
    def wrapper(*args, **kwargs):
        caller_specified_fields = dict(f_sig.bind(*args, **kwargs).arguments)
        self = caller_specified_fields.pop('self')

        f_defaults = {k: v.default for k, v in \
                       signature(f).parameters.items() \
                       if not v.default is _empty}

        f_defaults.update(caller_specified_fields)
        # Set default values
        for k, v in f_defaults.items():
            setattr(self, k, v)

        return f(*args, **kwargs)

    return wrapper


def run_super_first(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        super_f = getattr(super(type(self), self), f.__name__)
        super_f(*args, **kwargs)
        return f(self)

    return wrapper


def fst(tup):
    return tup[0]


def snd(tup):
    return tup[1]


def swap(tup):
    x, y = tup
    return y, x


def f_apply(f, x):
    return f(x)


def assert_len(l, xs):
    # TODO: custom exception classes
    assert len(xs) == l, "List did not have the expected number of items!"
    return xs


def compose(f,g):
    return lambda x: f(g(x))


def intercalate(x, ys):
    return x.join([x for x in ys if x])


#NOTE: This is a decorator function, not a class!
class add_methods(object):
    '''Add the public attributes of a mixin to another class.
    Attribute name collisions result in a TypeError if force is False.
    If a mixin is an ABC, the decorated class is registered to it,
    indicating that the class implements the mixin's tooling.
    '''
    @auto_set_fields
    def __init__(self, *attrs):
        pass

    def __call__(self, cls):
        for attr in self.attrs:
            if hasattr(cls, attr.__name__):
                raise TypeError("name collision ({})".format(attr.__name__))
            else:
                setattr(cls, attr.__name__, attr)

        return cls


#NOTE: This is a decorator function, not a class!
class add_method_as(object):
    '''
    Add the public attributes of a mixin to another class.
    Attribute name collisions result in a TypeError if force is False.
    If a mixin is an ABC, the decorated class is registered to it,
    indicating that the class implements the mixin's tooling.
    '''
    @auto_set_fields
    def __init__(self, attr, as_name):
        pass

    def __call__(self, cls):
        if hasattr(cls, self.as_name):
            raise TypeError("name collision ({})".format(self.as_name))
        else:
            setattr(cls, self.as_name, self.attr)

        return cls


def filter_split(f, l):
    accepted = []
    rejected = []
    [accepted.append(x) if f(x) else rejected.append(x) for x in l]
    return accepted, rejected


def dict_to_nt(d, typename='seneca_generate_type'):
    return namedtuple(typename, d.keys())(*d.values())


def dict_to_obj(d, typename=None):
    empty = lambda: None

    for k,v in d.items():
        setattr(empty, k, v)

    return empty


def manual_import(path, name):
    '''
    TODO: More tests.
    '''
    with open(path, 'r') as file:
        module_string = file.read()

    mod_dict = {'__file__': path, '__name__': name}
    mod_ast = ast.parse(module_string)
    mod_comp = compile(mod_ast, filename=name, mode="exec")
    exec(mod_comp, mod_dict)

    return mod_dict



from seneca.engine.conflict_resolution import *
from seneca.engine.cr_commands import *
import redis
from unittest import TestCase
import unittest


class TestCRGetSet(TestCase):

    def setUp(self):
        self.master = redis.StrictRedis(host='localhost', port=6379, db=0)
        self.working = redis.StrictRedis(host='localhost', port=6379, db=1)
        self.sbb_data = {}

    def tearDown(self):
        self.master.flushdb()
        self.working.flushdb()

    def _new_getset(self, should_set=True, sbb_idx=0, contract_idx=0, finalize=False):
        if contract_idx in self.sbb_data:
            data = self.sbb_data[contract_idx]
        else:
            data = self._new_cr_data(sbb_idx=sbb_idx, finalize=finalize)
            self.sbb_data[contract_idx] = data

        if should_set:
            return CRCmdSet(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, contract_idx=contract_idx,
                            data=data)
        else:
            return CRCmdGet(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, contract_idx=contract_idx,
                            data=data)

    def _new_cr_data(self, sbb_idx=0, finalize=False):
        return CRContext(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, finalize=finalize)

    def _new_get(self, sbb_idx=0, contract_idx=0, finalize=False):
        return self._new_getset(should_set=False, sbb_idx=0, contract_idx=contract_idx, finalize=finalize)

    def _new_set(self, sbb_idx=0, contract_idx=0, finalize=False):
        return self._new_getset(should_set=True, sbb_idx=0, contract_idx=contract_idx, finalize=finalize)

    def test_get_from_master(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE)

        actual = cr_get(KEY)

        self.assertEqual(actual, VALUE)

    def test_get_from_common(self):
        KEY = 'im_a_key'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE_M)
        self.working.set(KEY, VALUE_C)

        actual = cr_get(KEY)

        self.assertEqual(actual, VALUE_C)

    def test_get_from_sbb_specific_original(self):
        KEY = 'im_a_key'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        VALUE_SBB = b'value_on_sbb'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE_M)
        self.working.set(KEY, VALUE_C)
        cr_get.data['getset'][KEY] = {'og': VALUE_SBB, 'mod': None, 'contracts': {0}}

        actual = cr_get(KEY)

        self.assertEqual(actual, VALUE_SBB)

    def test_get_from_sbb_specific_modified(self):
        KEY = 'im_a_key'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        VALUE_SBB_OG = b'value_on_sbb_og'
        VALUE_SBB_MOD = b'value_on_sbb_mod'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE_M)
        self.working.set(KEY, VALUE_C)
        cr_get.data['getset'][KEY] = {'og': VALUE_SBB_OG, 'mod': VALUE_SBB_MOD, 'contracts': {0}}

        actual = cr_get(KEY)

        self.assertEqual(actual, VALUE_SBB_MOD)

    def test_get_copies_original_from_master(self):
        KEY = 'im_a_key'
        VALUE_M = b'value_on_master'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE_M)

        cr_get(KEY)  # calling get should trigger the key to be copied to the SBB specific layer

        self.assertTrue(cr_get._sbb_original_exists(KEY))
        self.assertEqual(cr_get.data['getset'][KEY], {'og': VALUE_M, 'mod': None, 'contracts': {0}})

    def test_get_copies_original_from_common(self):
        KEY = 'im_a_key'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE_M)
        self.working.set(KEY, VALUE_C)

        cr_get(KEY)  # calling get should trigger the key to be copied to the SBB specific layer

        self.assertTrue(cr_get._sbb_original_exists(KEY))

        actual = cr_get(KEY)
        self.assertEqual(VALUE_C, actual)

    def test_basic_set(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set(KEY, NEW_VALUE)

        expected = {'og': VALUE, 'mod': NEW_VALUE, 'contracts': {0}}
        self.assertEqual(expected, cr_set.data['getset'][KEY])

    def test_basic_set_adds_to_redo_log(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set(KEY, NEW_VALUE)

        self.assertEqual(cr_set.data['getset'].redo_log[0][KEY], VALUE)

    def test_basic_set_new_key_adds_to_redo_log(self):
        KEY = 'im_a_key'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()

        cr_set(KEY, NEW_VALUE)

        self.assertEqual(cr_set.data['getset'].redo_log[0][KEY], None)

    def test_basic_set_adds_to_writes(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set(KEY, NEW_VALUE)

        writes = cr_set.data['getset'].writes
        self.assertTrue(KEY in writes[0])

    def test_basic_set_adds_to_key_contract_mods(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set(KEY, NEW_VALUE)

        contracts_mod = cr_set.data['getset'][KEY]['contracts']
        self.assertTrue(cr_set.contract_idx in contracts_mod)

    def test_basic_get_adds_to_reads(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        cr_get = self._new_get()
        self.master.set(KEY, VALUE)

        actual = cr_get(KEY)
        self.assertEqual(actual, VALUE)

        reads = cr_get.data['getset'].reads
        self.assertTrue(KEY in reads[0])

    def test_basic_set_and_rollback(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        cr_set = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set(KEY, NEW_VALUE)

        cr_set.data['getset'].rollback_contract(0)
        expected = {'og': VALUE, 'mod': VALUE, 'contracts': set()}
        self.assertEqual(expected, cr_set.data['getset'][KEY])

    def test_basic_set_and_rollback_two_contracts(self):
        KEY = 'im_a_key'
        VALUE = b'value_on_master'
        NEW_VALUE = b'new_value'
        NEWER_VALUE = b'newer_value'
        cr_set1 = self._new_set()
        self.master.set(KEY, VALUE)

        cr_set1(KEY, NEW_VALUE)

        cr_set2 = self._new_set(contract_idx=1)
        cr_set2(KEY, NEWER_VALUE)

        cr_set2.data['getset'].rollback_contract(1)
        expected = {'og': VALUE, 'mod': NEW_VALUE, 'contracts': {0}}
        self.assertEqual(expected, cr_set1.data['getset'][KEY])

    def test_basic_setget_and_reset_contract_data(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        cr_get = self._new_get()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_get(KEY2)
        cr_set(KEY1, NEW_VALUE1)

        cr_get.data['getset'].reset_contract_data(0)
        self.assertEqual(len(cr_set.data['getset'].writes[0]), 0)
        self.assertEqual(len(cr_set.data['getset'].reads[0]), 0)

    def test_adds_key_that_does_not_yet_exist(self):
        KEY = 'im_a_key'
        VALUE = b'g00d_val'
        cr_set = self._new_set()

        cr_set(KEY, VALUE)

        expected = {'og': None, 'mod': VALUE, 'contracts': {0}}
        self.assertEqual(expected, cr_set.data['getset'][KEY])

    def test_should_rerun_no_changes(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        cr_get = self._new_get()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_get(KEY2)
        cr_set(KEY1, NEW_VALUE1)

        cr_data = cr_set.data['getset']
        self.assertFalse(0 in list(cr_data.get_rerun_list(reset_keys=False)))

    def test_should_rerun_change_common_only_write(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to common on the written key
        self.working.set(KEY1, b'new_common1')

        cr_data = cr_set.data['getset']
        self.assertTrue(0 in list(cr_data.get_rerun_list(reset_keys=False)))


    def test_should_rerun_change_common_only_read(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        cr_get = self._new_get()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_get(KEY1)

        # Make some changes to common on the written key
        self.working.set(KEY1, b'new_common1')

        cr_data = cr_get.data['getset']
        self.assertTrue(0 in list(cr_data.get_rerun_list(reset_keys=False)))

    def test_should_rerun_change_master_only_read(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        cr_get = self._new_get()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_get(KEY1)

        # Make some changes to common on the written key
        self.master.set(KEY1, b'new_common1')

        cr_data = cr_get.data['getset']
        self.assertTrue(0 in list(cr_data.get_rerun_list(reset_keys=False)))

    def test_should_rerun_change_master_only_write(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to common on the written key
        self.master.set(KEY1, b'new_common1')

        cr_data = cr_set.data['getset']
        self.assertTrue(0 in list(cr_data.get_rerun_list(reset_keys=False)))

    def test_get_modified_keys_from_common(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to common on the written key
        self.working.set(KEY1, b'new_common1')

        cr_data = cr_set.data['getset']
        actual_mods = cr_data.get_modified_keys()
        expected_mods = {KEY1}
        self.assertEqual(expected_mods, actual_mods)

    def test_get_modified_keys_from_master(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to master on the written key
        self.master.set(KEY1, b'new_common1')

        cr_data = cr_set.data['getset']
        actual_mods = cr_data.get_modified_keys()
        expected_mods = {KEY1}
        self.assertEqual(expected_mods, actual_mods)

    def test_reset_key_from_common(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        COMMON_VALUE1 = b'new_common1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to common on the written key
        self.working.set(KEY1, COMMON_VALUE1)

        cr_data = cr_set.data['getset']
        cr_data.reset_key(KEY1)

        self.assertEqual(cr_data[KEY1]['og'], COMMON_VALUE1)
        self.assertEqual(cr_data[KEY1]['mod'], None)

    def test_reset_key_from_master(self):
        KEY1 = 'im_a_key1'
        KEY2 = 'im_a_key2'
        VALUE1 = b'value_on_master1'
        VALUE2 = b'value_on_master2'
        NEW_VALUE1 = b'new_value1'
        COMMON_VALUE1 = b'new_common1'
        MASTER_VALUE1 = b'new_master1'
        cr_set = self._new_set()
        self.master.set(KEY1, VALUE1)
        self.master.set(KEY2, VALUE2)

        cr_set(KEY1, NEW_VALUE1)

        # Make some changes to master on the written key
        self.master.set(KEY1, MASTER_VALUE1)

        cr_data = cr_set.data['getset']
        cr_data.reset_key(KEY1)

        self.assertEqual(cr_data[KEY1]['og'], MASTER_VALUE1)
        self.assertEqual(cr_data[KEY1]['mod'], None)

    def test_get_modified_keys_recursive(self):
        k = ['k{}'.format(i+1) for i in range(5)]
        v = ['v{}'.format(i+1) for i in range(5)]

        cr_set = self._new_set()
        cr_data = cr_set.data['getset']

        for i in range(len(k)):
            cr_data[k[i]] = {'og': v[i], 'mod': v[i] + '_MOD', 'contracts': set()}

        cr_data[k[0]]['contracts'] = {0, 2}
        cr_data[k[1]]['contracts'] = {0, 3}
        cr_data[k[2]]['contracts'] = {1}
        cr_data[k[3]]['contracts'] = {2}
        cr_data[k[4]]['contracts'] = {3}

        cr_data.writes[0] = {k[0], k[1]}
        cr_data.reads[0] = {k[1]}

        cr_data.writes[1] = {k[2]}
        cr_data.reads[1] = {k[2]}

        cr_data.writes[2] = {k[0], k[3]}
        cr_data.reads[2] = {k[0]}

        cr_data.writes[3] = {k[1], k[4]}
        cr_data.reads[3] = {k[1]}

        self.master.set(k[3], 'SOME NEW VALUE FOR k[3]')

        expected_mods = {k[3], k[0], k[1], k[4]}
        actual_mods = cr_data.get_modified_keys_recursive()
        self.assertEqual(expected_mods, actual_mods)


if __name__ == "__main__":
    unittest.main()



import nacl
import nacl.encoding
import nacl.signing


def verify(v: bytes, msg: bytes, sig: bytes):
    v = nacl.signing.VerifyKey(v)
    try:
        v.verify(msg, sig)
    except nacl.exceptions.BadSignatureError:
        return False
    return True

def run_tests(_):
    import doctest, sys
    #return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})

    from seneca.execute import Empty

    res = Empty()
    res.attempted = 0
    res.failed = 0

    s = nacl.signing.SigningKey.generate()
    msg = b'howdy partner'

    sig = s.sign(msg)

    # pynacl includes the message in the signature, which we are not interested in
    sig = sig[:-len(msg)]

    v = s.verify_key.encode()

    if verify(v, msg, sig):
        res.attempted = 1
    else:
        res.failed = 1

    return res


from unittest import TestCase
from seneca.engine.book_keeper import BookKeeper
from unittest.mock import patch
from multiprocessing import Process
import threading, unittest
from seneca.engine.conflict_resolution import CRContext


class TestBookKeeper(TestCase):

    def _build_info_dict(self, sbb_idx=0, contract_idx=0, master_db='', working_db=''):
        master_db = master_db or 'some placeholder that irl would be a redis client cursor'
        working_db = working_db or 'another placeholder that irl would be a redis client cursor'
        data = CRContext(working_db=working_db, master_db=master_db, sbb_idx=sbb_idx, finalize=False)

        info = {'sbb_idx': sbb_idx, 'contract_idx': contract_idx, 'data': data}
        return info

    def test_set_and_get_info(self):
        expected_info = self._build_info_dict(sbb_idx=8, contract_idx=10)

        BookKeeper.set_info(**expected_info)
        actual_info = BookKeeper.get_info()

        self.assertEqual(actual_info, expected_info)

    def test_set_and_get_multiproc(self):
        # TODO -- implement ... spin up multiple theads/procs and ensure each has a different ID. Will have to write
        # to some shared Queue object so that the assertions can be checked on the main process (the one running this test)
        pass
        # with patch('os.getpid') as pid_func:
        #     with patch('threading.get_ident') as id_func:
        #         mock_pid = 100
        #         mock_thread = 90
        #         pid_func.return_value = mock_pid
        #         id_func.return_value = mock_thread
        #
        #         print("Got key: {}".format(BookKeeper._get_key()))


if __name__ == '__main__':
    unittest.main()



import unittest, os, sys, importlib
from os import getenv as env

testmodules = [
    'tests.unit.interface',
    'tests.unit.engine',
    'tests.unit.security',
    'tests.unit.contracts',
]

loader = unittest.TestLoader()

for t in testmodules:
    try:
        # If the module defines a suite() function, call it to get the suite.
        suite = loader.discover(t)
        mod = __import__(t, globals(), locals(), ['suite'])
        suitefn = getattr(mod, 'suite')
        suite.addTest(suitefn())
    except (ImportError, AttributeError):
        # else, just load all the test cases from the module.
        suite.addTest(unittest.defaultTestLoader.loadTestsFromName(t))

    result = unittest.TextTestRunner().run(suite)
    if not result.wasSuccessful():
        sys.exit(1)



from unittest import TestCase
from seneca.engine.util import *
import unittest

class TestUtil(TestCase):
    def test_fst(self):
        self.assertEqual(fst((1, 2)), 1)

    def test_snd(self):
        self.assertEqual(snd((1, 2)), 2)

    def test_swap(self):
        self.assertEqual(swap((1, 2)), (2, 1))

    def test_f_apply(self):
        self.assertEqual(f_apply(lambda x: x, 1), 1)

    def test_intercalate(self):
        self.assertEqual(intercalate('_', 'abcde'), 'a_b_c_d_e')
        self.assertEqual(intercalate('_', [None, 'a', None, 'b']), 'a_b')

    def test_add_methods(self):
        def id_(self, x):
            return x

        @add_methods(id_)
        class Test(object):
            pass

        t = Test()
        self.assertEqual(t.id_('abc'), 'abc')

    def test_add_methods_as(self):
        def id_(self, x):
            return x

        @add_method_as(id_, 'id_')
        class Test(object):
            pass

        t = Test()
        self.assertEqual(t.id_('abc'), 'abc')

    def test_filter_split(self):
        self.assertEqual(filter_split(lambda x: x == 'x', 'aaxbbxccxddx'),
                         (['x', 'x', 'x', 'x'], ['a', 'a', 'b', 'b', 'c', 'c', 'd', 'd']))

    def test_dict_to_nt(self):
        d = dict_to_nt({'x': 'y'})
        self.assertEqual(str(d), "seneca_generate_type(x='y')")

    def test_dict_to_obj(self):
        self.assertEqual(str(dict_to_obj({'x': 'y'}).x), 'y')

    def test_manual_import(self):
        m = manual_import(__file__, 'tester')
        self.assertEqual(type(m), dict)
        self.assertTrue('manual_import' in m.keys())



if __name__ == '__main__':
    unittest.main()



from seneca.engine.interface import SenecaInterface
import types

class SenecaFunction:
    def __init__(self, name, module_path, kwargs, default_sender, driver):
        self.name = name
        self.module_path = module_path
        self.kwargs = kwargs
        self.defaults = {
            'sender': default_sender
        }
        self.driver = driver

    def __call__(self, *args, **kwargs):

        def default(d, k):
            return d if kwargs.get(k) is None else kwargs.get(k)

        stamps = default(None, 'stamps')
        sender = default(self.defaults.get('sender'), 'sender')

        r = self.driver.execute_function(
            module_path=self.module_path,
            stamps= stamps,
            sender= sender,
            **kwargs
        )

        return r


class ContractWrapper:
    def __init__(self, contract_name=None, driver=None, default_sender=None):
        self.driver = driver
        self.author = driver.get_contract_meta(contract_name)['author']
        self.default_sender = default_sender
        contract_code = driver.get_code_obj(contract_name)

        codes = [cd for cd in contract_code.co_consts if type(cd) == types.CodeType]
        for _c in codes:
            name = _c.co_name
            module_path = 'seneca.contracts.{}.{}'.format(contract_name, _c.co_name)
            kwargs = _c.co_varnames
            setattr(self, name, SenecaFunction(name=name,
                                               module_path=module_path,
                                               kwargs=kwargs,
                                               default_sender=self.default_sender,
                                               driver=self.driver))



import ast, builtins

ALLOWED_AST_TYPES = {
    ast.Module,
    ast.Eq,
    ast.Call,
    ast.Dict,
    ast.Attribute,
    ast.Pow,
    ast.Index,
    ast.Not,
    ast.alias,
    ast.If,
    ast.FunctionDef,
    ast.GtE,
    ast.LtE,
    ast.Load,
    ast.arg,
    ast.Add,
    ast.Lambda,
    ast.Import,
    ast.ImportFrom,
    ast.Name,
    ast.Num,
    ast.BinOp,
    ast.Store,
    ast.Assert,
    ast.Assign,
    ast.AugAssign,
    ast.Subscript,
    ast.Compare,
    ast.Return,
    ast.NameConstant,
    ast.Expr,
    ast.keyword,
    ast.Sub,
    ast.arguments,
    ast.List,
    ast.Set,
    ast.Str,
    ast.UnaryOp,
    ast.Pass,
    ast.Tuple,
    ast.Div,
    ast.In,
    ast.NotIn,
    ast.Gt,
    ast.Lt,
    ast.Starred,
    ast.Mod,
    ast.NotEq,
    # TODO: Decide if we actually want these
    # Error handling
    # ast.ExceptHandler,
    # ast.Try,
    # Loops
    ast.For,
    ast.While,
    # comprehension
    ast.ListComp,
    ast.comprehension,
    ast.Slice,
    ast.USub,
    # Conditonals
    ast.BoolOp,
    ast.And,
    ast.Or,
    ast.Mult,
}

SENECA_LIBRARY_PATH = 'seneca.libs'

ALLOWED_IMPORT_PATHS = [
    'seneca.contracts',
    'test_contracts'
]

_SAFE_NAMES = [
    '__import__',

    'None',
    'False',
    'True',

    'callable',
    'isinstance',
    'issubclass',

    'abs',
    'bool',
    'chr',
    'complex',
    'divmod',
    'float',
    'hash',
    'hex',
    'id',
    'int',

    'len',
    'oct',
    'ord',
    'pow',
    'range',
    'repr',
    'round',
    'slice',
    'str',
    'tuple',
    'zip',

    # JUST FOR TESTING! TODO remove this irl
    'dir',
    'help',
    'print',
    # 'globals'
]

_SAFE_EXCEPTIONS = [
    'ArithmeticError',
    'AssertionError',
    'AttributeError',
    'BaseException',
    'BufferError',
    'BytesWarning',
    'DeprecationWarning',
    'EOFError',
    'EnvironmentError',
    'Exception',
    'FloatingPointError',
    'FutureWarning',
    'GeneratorExit',
    'IOError',
    'ImportError',
    'ImportWarning',
    'IndentationError',
    'IndexError',
    'KeyError',
    'KeyboardInterrupt',
    'LookupError',
    'MemoryError',
    'NameError',
    'NotImplementedError',
    'OSError',
    'OverflowError',
    'PendingDeprecationWarning',
    'ReferenceError',
    'RuntimeError',
    'RuntimeWarning',
    'StopIteration',
    'SyntaxError',
    'SyntaxWarning',
    'SystemError',
    'SystemExit',
    'TabError',
    'TypeError',
    'UnboundLocalError',
    'UnicodeDecodeError',
    'UnicodeEncodeError',
    'UnicodeError',
    'UnicodeTranslateError',
    'UnicodeWarning',
    'UserWarning',
    'ValueError',
    'Warning',
    'ZeroDivisionError',
]

SAFE_BUILTINS = {}

for name in _SAFE_NAMES:
    SAFE_BUILTINS[name] = getattr(builtins, name)

for name in _SAFE_EXCEPTIONS:
    SAFE_BUILTINS[name] = getattr(builtins, name)



from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter, ReadOnlyException, CompilationException
from os.path import join
from tests.utils import captured_output, TestInterface
import redis, unittest, seneca

test_contracts_path = seneca.__path__[0] + '/test_contracts/'

class TestSubmission(TestInterface):

    def test_publish_code_str(self):
        """
            Testing to see if the submission to Redis works.
        """
        code_str = """
@export
def ok():
    print('i am fine')
        """
        self.si.publish_code_str('crazy', 'anonymoose', code_str)
        self.si.execute_code_str("""
from seneca.contracts.crazy import ok
ok()
        """)
        self.assertEqual(code_str, self.si.get_code_str('crazy'))

    def test_publish_bad_code(self):
        """
            Trying to import protected functions will fail
        """
        code_str = """
from test_contracts.good import one_you_cannot_export
        """
        with self.assertRaises(CompilationException) as context:
            self.si.publish_code_str('incorrect', 'anonymoose', code_str)

    def test_publish_bad_code_inside_function(self):
        """
            Cannot import protected code inside a function neither.
        """
        code_str = """
def bad_code():
    from test_contracts.good import one_you_cannot_export
        """
        with self.assertRaises(CompilationException) as context:
            self.si.publish_code_str('incorrect', 'anonymoose', code_str)

    def test_republish_code_str_fail(self):
        """
            Republishting code to the same smart contract name will fail
        """
        self.si.publish_code_str('crazy', 'anonymoose', """
def ok():
    print('i am fine')
        """)
        with self.assertRaises(Exception) as context:
            self.si.publish_code_str('crazy', 'anonymoose', """
def fail():
    print('i am not fine')
            """)


if __name__ == '__main__':
    unittest.main()



from seneca.engine.conflict_resolution import *
from seneca.engine.cr_commands import *
import redis
from unittest import TestCase
import unittest


class TestCRHMap(TestCase):

    def setUp(self):
        self.master = redis.StrictRedis(host='localhost', port=6379, db=0)
        self.working = redis.StrictRedis(host='localhost', port=6379, db=1)
        self.sbb_data = {}

    def tearDown(self):
        self.master.flushdb()
        self.working.flushdb()

    def _new_getset(self, should_set=True, sbb_idx=0, contract_idx=0, finalize=False):
        if contract_idx in self.sbb_data:
            data = self.sbb_data[contract_idx]
        else:
            data = self._new_cr_data(sbb_idx=sbb_idx, finalize=finalize)
            self.sbb_data[contract_idx] = data

        if should_set:
            return CRCmdHSet(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, contract_idx=contract_idx,
                             data=data)
        else:
            return CRCmdHGet(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, contract_idx=contract_idx,
                             data=data)

    def _new_cr_data(self, sbb_idx=0, finalize=False):
        return CRContext(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, finalize=finalize)

    def _new_get(self, sbb_idx=0, contract_idx=0, finalize=False):
        return self._new_getset(should_set=False, sbb_idx=0, contract_idx=contract_idx, finalize=finalize)

    def _new_set(self, sbb_idx=0, contract_idx=0, finalize=False):
        return self._new_getset(should_set=True, sbb_idx=0, contract_idx=contract_idx, finalize=finalize)

    def test_get_from_master(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE = b'value_on_master'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE)

        actual = cr_get(KEY, FIELD)

        self.assertEqual(actual, VALUE)
        self.assertEqual({'og': VALUE, 'mod': None}, cr_get.data['hm'][KEY][FIELD])

    def test_get_from_common(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE_M)
        self.working.hset(KEY, FIELD, VALUE_C)

        actual = cr_get(KEY, FIELD)

        self.assertEqual(actual, VALUE_C)
        self.assertEqual({'og': VALUE_C, 'mod': None}, cr_get.data['hm'][KEY][FIELD])

    def test_get_from_sbb_specific_original(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        VALUE_SBB = b'value_on_sbb'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE_M)
        self.working.hset(KEY, FIELD, VALUE_C)
        cr_get.data['hm'][KEY][FIELD] = {'og': VALUE_SBB, 'mod': None}

        actual = cr_get(KEY, FIELD)

        self.assertEqual(actual, VALUE_SBB)

    def test_get_from_sbb_specific_modified(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        VALUE_SBB_OG = b'value_on_sbb_og'
        VALUE_SBB_MOD = b'value_on_sbb_mod'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE_M)
        self.working.hset(KEY, FIELD, VALUE_C)
        cr_get.data['hm'][KEY][FIELD] = {'og': VALUE_SBB_OG, 'mod': VALUE_SBB_MOD}

        actual = cr_get(KEY, FIELD)

        self.assertEqual(actual, VALUE_SBB_MOD)

    def test_get_copies_original_from_master(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE_M = b'value_on_master'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE_M)

        cr_get(KEY, FIELD)  # calling get should trigger the key to be copied to the SBB specific layer

        self.assertTrue(cr_get._sbb_original_exists(KEY, FIELD))
        self.assertEqual(cr_get.data['hm'][KEY][FIELD], {'og': VALUE_M, 'mod': None})

    def test_get_copies_original_from_common(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE_M = b'value_on_master'
        VALUE_C = b'value_on_common'
        cr_get = self._new_get()
        self.master.hset(KEY, FIELD, VALUE_M)
        self.working.hset(KEY, FIELD, VALUE_C)

        cr_get(KEY, FIELD)  # calling get should trigger the key to be copied to the SBB specific layer

        self.assertTrue(cr_get._sbb_original_exists(KEY, FIELD))

        actual = cr_get(KEY, FIELD)
        self.assertEqual(VALUE_C, actual)

    # TODO fix this test once we implement write log for hash maps
    # def test_basic_set(self):
    #     KEY = 'im_a_key'
    #     FIELD = 'im_a_field'
    #     VALUE = b'value_on_master'
    #     NEW_VALUE = b'new_value'
    #     cr_set = self._new_set()
    #     cr_get = self._new_get()
    #     self.master.hset(KEY, FIELD, VALUE)
    #     cr_set(KEY, FIELD, NEW_VALUE)
    #
    #     expected = {'og': VALUE, 'mod': NEW_VALUE}
    #     expected_mod = cr_set._get_key_field_name(KEY, FIELD)
    #     actual_mods = cr_set.data['hm'].mods
    #     self.assertEqual(expected, cr_set.data['hm'][KEY][FIELD])
    #     self.assertEqual(expected['mod'], cr_get(KEY, FIELD))
    #
    #     self.assertEqual(len(actual_mods), 1)
    #     self.assertEqual(actual_mods[0], {expected_mod})

    def test_adds_key_that_does_not_yet_exist(self):
        KEY = 'im_a_key'
        FIELD = 'im_a_field'
        VALUE = b'g00d_val'
        cr_set = self._new_set()

        cr_set(KEY, FIELD, VALUE)

        expected = {'og': None, 'mod': VALUE}
        self.assertEqual(expected, cr_set.data['hm'][KEY][FIELD])


if __name__ == "__main__":
    unittest.main()




import redis
import ujson as json
from seneca.constants.config import get_redis_port, MASTER_DB, DB_OFFSET, get_redis_password
from seneca.libs.logger import get_logger
from seneca.engine.book_keeper import BookKeeper
from seneca.engine.interpreter import SenecaInterpreter, Seneca
from seneca.engine.conflict_resolution import RedisProxy
from decimal import Decimal
from seneca.libs.decimal import make_decimal

REDIS_PORT = get_redis_port()
REDIS_PASSWORD = get_redis_password()

'''

Datatype serialization format:

type<prefix>(declaration)

:map(str, int)
:map<coins>(str, int)

:list<todo>(:map(str, int))


'''

COMPLEX_TYPE_PREFIX = '*'
CTP = COMPLEX_TYPE_PREFIX

# # # CLEAN THIS UP! GOOD LORD!!!
type_to_string = {
    str: 'str',
    int: 'int',
    bool: 'bool',
    bytes: 'bytes',
    float: 'float',
    Decimal: 'float'
}


string_to_type = {
    'str': str,
    'int': int,
    'bool': bool,
    'bytes': bytes,
    'float': Decimal
}

vivified_primitives = {
    int: 0,
    str: '',
    bool: False,
    float: Decimal('0'),
    bytes: b''
}


primitive_types = [int, str, bool, bytes, Decimal, float, None]


def extract_prefix(s):
    if s[0] == '<':
        prefix_idx_end = s.find('>')
        prefix_s = s[:prefix_idx_end]
        return prefix_s.split(':')[-1], s[1+prefix_idx_end:]
    return None, s

'''
Returns the representation of the complex type if it is not a primative.
Otherwise, returns
'''


def encode_type(t):
    if isinstance(t, RObject) or isinstance(t, Placeholder):
        return t.rep()
    return type_to_string.get(t)


complex_tokens = ['map', 'list', 'table', 'ranked']
all_tokens = ['int', 'str', 'bool', 'bytes', 'float', 'map', 'list', 'table', 'ranked']
# # #


def parse_representation(s):
    if s[0] == CTP:
        return parse_complex_type_repr(s)
    else:
        return string_to_type.get(s)


def parse_type_repr(s):
    if s in complex_tokens and s[0] == CTP:
        return parse_complex_type_repr(s)
    elif s in string_to_type.keys():
        return string_to_type.get(s)
    return None


def parse_complex_type_repr(s):
    s = s[1:]
    for t in complex_tokens:
        if s.startswith(t):
            if t == 'table':
                return build_table_from_repr(s)
            elif t == 'list':
                return build_list_from_repr(s)
            elif t == 'map':
                return build_map_from_repr(s)
            elif t == 'ranked':
                return build_ranked_from_repr(s)


def build_table_from_repr(s):
    t = {}

    slice_idx = s.find('table') + len('table')
    s = s[slice_idx:]

    # check if the prefix has been defined
    prefix, s = extract_prefix(s)

    start = s.find('({') + 2
    end = s.find('})')

    s = s[start:end]

    while len(s) > 0:

        key_end = s.find(':')
        key = s[:key_end]

        s = s[key_end:]

        # whichever index is lower is the next value
        # hi:int, ...
        # hi:map(str, ...
        next_simple_type = s.find(',')
        next_complex_type = s.find('(')

        true_next_type = next_simple_type if next_simple_type < next_complex_type else next_complex_type

        if next_simple_type < next_complex_type:
            value = s[1:true_next_type]
            t[key] = string_to_type.get(value)
            s = s[1 + true_next_type:]

        else:
            value_idx_end = s.find(')', next_complex_type)
            value = s[1:value_idx_end + 1]
            t[key] = parse_complex_type_repr(value)
            s = s[1 + value_idx_end + 1:]

    if prefix is not None:
        return table(prefix=prefix, schema=t)
    return TablePlaceholder(schema=t)


def build_list_from_repr(s):
    slice_idx = s.find('list') + len('list')
    s = s[slice_idx:]

    # check if the prefix has been defined
    prefix, s = extract_prefix(s)

    _type = s[1:-1]

    value_type = parse_type_repr(_type)

    if prefix is not None:
        return hlist(prefix=prefix, value_type=value_type)
    return ListPlaceholder(value_type=value_type)


def build_map_from_repr(s):
    slice_idx = s.find('map') + len('map')
    s = s[slice_idx:]

    # check if the prefix has been defined
    prefix, s = extract_prefix(s)

    types = s.split(',')

    assert len(types) == 2, 'Too many types provided to the map representation string! {}'.format(types)

    key_type = parse_type_repr(types[0][1:])
    value_type = parse_type_repr(types[1][:-1])

    if prefix is not None:
        return hmap(prefix=prefix, key_type=key_type, value_type=value_type)
    return Placeholder(key_type=key_type, value_type=value_type, placeholder_type=HMap)


def build_ranked_from_repr(s):
    slice_idx = s.find('ranked') + len('ranked')
    s = s[slice_idx:]

    # check if the prefix has been defined
    prefix, s = extract_prefix(s)

    types = s.split(',')

    assert len(types) == 2, 'Too many types provided to the map representation string! {}'.format(types)

    key_type = parse_type_repr(types[0][1:])
    value_type = parse_type_repr(types[1][:-1])

    return ranked(prefix=prefix, key_type=key_type, value_type=value_type)


class Placeholder:
    def __init__(self, key_type=str, value_type=int, placeholder_type=None):
        assert placeholder_type is not None and type(placeholder_type) == type, 'Provide a type to represent.'
        self.key_type = key_type
        self.value_type = value_type
        self.placeholder_type = placeholder_type

    def valid(self, t):
        if self.key_type == t.key_type and \
                self.value_type == t.value_type and \
                type(t) == self.placeholder_type:
            return True
        return False

    def rep(self):
        return CTP + 'map' + '(' + encode_type(self.key_type) + ',' + encode_type(self.value_type) + ')'


class ListPlaceholder(Placeholder):
    def __init__(self, value_type=int):
        self.key_type = str
        self.value_type = value_type
        self.placeholder_type = HList

    def valid(self, t):
        if self.value_type == t.value_type and type(t) == self.placeholder_type:
            return True
        return False

    def rep(self):
        return CTP + 'list' + '(' + encode_type(self.value_type) + ')'


class TablePlaceholder(Placeholder):
    def __init__(self, key_type=str, schema=None):
        self.key_type = key_type
        self.schema = schema
        self.placeholder_type = Table

    def valid(self, t):
        self_keys = set(self.schema.keys())
        t_keys = set(t.schema.keys())

        if self_keys == t_keys and self.key_type == t.key_type \
                and t.prefix is not None and isinstance(t, Table):
            return True
        return False

    def rep(self):
        d = '({'
        for k, v in self.schema.items():
            d += '{}'.format(k)
            d += ':'
            d += encode_type(v)
            d += ','
        d = d[:-1]
        d += '})'
        return CTP + 'table' + d


class RankedPlaceholder(Placeholder):
    def __init__(self, key_type=str, value_type=int):
        self.key_type = str
        self.value_type = value_type
        self.placeholder_type = Ranked

    def valid(self, t):
        if self.key_type == t.key_type and \
                self.value_type == t.value_type and \
                type(t) == self.placeholder_type:
            return True
        return False

    def rep(self):
        return CTP + 'ranked' + '(' + encode_type(self.key_type) + ',' + encode_type(self.value_type) + ')'


def is_complex_type(v):
    for t in complex_types:
        if (issubclass(type(v), Placeholder) and issubclass(v.placeholder_type, t)) \
                or issubclass(type(v), t):
            return True
    return False


# table to be done later
def vivify(potential_prefix, t):
    if t in primitive_types:
        return vivified_primitives[t]
    elif issubclass(type(t), Placeholder):
        if t.placeholder_type == HMap:
            return hmap(prefix=potential_prefix, key_type=t.key_type, value_type=t.value_type)
        elif t.placeholder_type == HList:
            return hlist(prefix=potential_prefix, value_type=t.value_type)
        #elif t == Table:
        #    return table(prefix=potential_prefix, key_type=t.key_type, schema=t.schema)
    elif t in complex_types:
        if type(t) == HMap:
            return hmap(prefix=potential_prefix, key_type=t.key_type, value_type=t.value_type)
        elif type(t) == HList:
            return hlist(prefix=potential_prefix, value_type=t.value_type)
        #elif type(t) == Table:
        #    return table(prefix=potential_prefix, key_type=t.key_type, schema=t.schema)
    return None


class RObject:

    def __init__(self, prefix=None, key_type=str, value_type=int, delimiter=':', rep_str='obj'):

        self.driver = Seneca.interface.r

        self.contract_id = Seneca.loaded['__main__']['rt']['contract']
        self.prefix = prefix
        self.concurrent_mode = Seneca.concurrent_mode
        self.key_type = key_type

        #self.driver = driver

        assert key_type is not None, 'Key type cannot be None'
        assert key_type in primitive_types or is_complex_type(key_type)

        # prevents you from requiring an RObject instance that has a prefix as a value type
        assert value_type in primitive_types or isinstance(value_type, Placeholder), \
            'You must pass a Placeholder object that does not contain a prefix as a value type. {}'.format(value_type)

        self.value_type = value_type

        self.delimiter = delimiter
        self.rep_str = rep_str

        if self.concurrent_mode:
            assert BookKeeper.has_info(), "No BookKeeping info found for this thread/process with key {}. Was set_info " \
                                          "called on this thread first?".format(BookKeeper._get_key())
            info = BookKeeper.get_info()
            self.driver = RedisProxy(sbb_idx=info['sbb_idx'], contract_idx=info['contract_idx'], data=info['data'])

    def encode_value(self, value, explicit=False):
        v = None

        if issubclass(type(self.value_type), Placeholder):
            assert self.value_type.valid(value) is True, \
                'Value {} is not a matching map'.format(value)
            v = value.rep()

        else:
            # due to the naive nature of fixed point precision casting, we try to cast decimals into ints when there
            # is no loss of precision

            if isinstance(value, Decimal) and \
               self.value_type == int and \
               value.quantize(Decimal(1)) == value:
                value = int(value)

            assert type(value) == self.value_type or \
                   self.value_type is None or \
                   explicit is True or \
                   (isinstance(value, Decimal) and self.value_type == float), \
                'Value is not of type "{}"'.format(self.value_type)
            if isinstance(value, float):
                v = make_decimal(value)
                v = str(v)
            else:
                v = json.dumps(value)

        v = v.encode()
        return v

    def decode_value(self, value):
        '''
        This is where that fun shit goes for 'parsing representation'
        '''
        if value is not None:
            value = value.decode()
            if value[0] == CTP:
                return parse_complex_type_repr(value)
            else:
                try:
                    value = Decimal(value)
                except:
                    value = json.loads(value)

        return value

    def check_key_type(self, key):
        msg = 'Key {} is not of type "{}"'.format(type(key), self.key_type)
        if type(self.key_type) == type:
            assert isinstance(key, self.key_type), msg
        else:
            assert self.key_type.valid(key), msg

    def rep(self):
        raise NotImplementedError

    def drop(self):
        return self.driver.delete(self.prefix)


####################
# HMap Datatype
####################
class HMap(RObject):
    def __init__(self, prefix=None,
                 key_type=str,
                 value_type=int
                 ):
        super().__init__(prefix=prefix,
                         key_type=key_type,
                         value_type=value_type,
                         rep_str='map')

        self.vivification_idx = 0

    def set(self, key, value):
        v = self.encode_value(value)
        self.check_key_type(key)

        if type(key) in complex_types:
            key = key.rep()
        return self.driver.set('{}{}{}'.format(self.prefix, self.delimiter, key), v)

    def get(self, key):
        if type(key) in complex_types:
            key = key.rep()

        g = self.driver.get('{}{}{}'.format(self.prefix, self.delimiter, key))
        g = self.decode_value(g)
        return g

    def __getitem__(self, k):
        item = self.get(k)
        if item is None and self.value_type is not None:
            return vivify('{}.{}'.format(self.prefix, k), self.value_type)
        return item

    def __setitem__(self, k, v):
        return self.set(k, v)

    def rep(self):
        return '{}map<{}:{}>({},{})'.format(CTP,
                                            self.contract_id,
                                            self.prefix,
                                            encode_type(self.key_type),
                                            encode_type(self.value_type))

    def exists(self, k):
        return self.driver.exists(k)

def hmap(prefix=None, key_type=str, value_type=int):
    if prefix is None:
        return Placeholder(key_type=key_type, value_type=value_type, placeholder_type=HMap)
    return HMap(prefix=prefix, key_type=key_type, value_type=value_type)

####################
# HList Datatype
####################
class HList(RObject):
    def __init__(self, prefix=None,
                 value_type=int
                 ):
        super().__init__(prefix=prefix,
                         key_type=str,
                         value_type=value_type,
                         rep_str='list')

    def get(self, i):
        g = self.driver.lindex(self.prefix, i)
        g = self.decode_value(g)
        return g

    def set(self, i, value):
        #TODO add error handling for trying to set indexes that don't exist
        v = self.encode_value(value)
        return self.driver.lset(self.prefix, i, v)

    def push(self, value):
        v = self.encode_value(value)
        return self.driver.lpush(self.prefix, v)

    def pop(self):
        g = self.driver.lpop(self.prefix)
        g = self.decode_value(g)
        return g

    def pop_right(self):
        g = self.driver.rpop(self.prefix)
        g = self.decode_value(g)
        return g

    def append(self, value):
        v = self.encode_value(value)
        return self.driver.rpush(self.prefix, v)

    def extend(self, l):
        for ll in l:
            self.append(ll)

    def __getitem__(self, i):
        item = self.get(i)
        if item is None and self.value_type is not None:
            return vivify('{}.{}'.format(self.prefix, i), self.value_type)
        return item

    def __setitem__(self, i, v):
        return self.set(i, v)

    def rep(self):
        return '{}list<{}:{}>({})'.format(CTP, self.contract_id, self.prefix, encode_type(self.value_type))

    def exists(self, k):
        return self.driver.exists(k)


def hlist(prefix=None, value_type=int):
    if prefix is None:
        return ListPlaceholder(value_type=value_type)
    return HList(prefix=prefix, value_type=value_type)


####################
# Table Datatype
####################
class Table(RObject):
    def __init__(self, prefix=None, key_type=str, schema=None):
        super().__init__(prefix=prefix,
                         key_type=key_type,
                         value_type=str,
                         rep_str='table')
        assert self.validate_schema(schema), 'Schema is not the correct type.'
        self.schema = schema
        self.p = self.prefix + self.delimiter

    def validate_schema(self, d):
        for k, v in d.items():
            if not isinstance(k, str):
                return False
            if not isinstance(v, type) and not isinstance(v, Placeholder):
                return False
        return True

    def dict_matches_schema(self, d):
        schema_set = set(self.schema.keys())
        d_set = set(d.keys())
        assert schema_set > d_set or schema_set == d_set, \
            'Mismatching keys. {} is not in {}'.format(d_set, schema_set)

        for k, v in d.items():
            assert type(v) == self.schema[k], \
                'Mismatching type. {} is not {}'.format(v, self.schema[k])

        return True

    def encode_value(self, value, t):
        v = None

        if issubclass(type(t), Placeholder):
            assert t.valid(value) is True, \
                'Value {} is not a matching map'.format(value)
            v = value.rep()
        else:
            assert type(value) == t or t is None, \
                'Value is not of type "{}"'.format(t)
            v = json.dumps(value)

        v = v.encode()
        return v

    def set(self, k, v):
        assert self.dict_matches_schema(v)
        self.check_key_type(k)

        # modify dictionary in place with encoded values
        for _k, _v in v.items():
            v[_k] = self.encode_value(_v, self.schema[_k])

        if type(k) in complex_types:
            k = k.rep()

        self.driver.hmset('{}{}'.format(self.p, k), v)

    def get(self, k, s=None):
        d = {}

        if s is None:
            keys = tuple(self.schema.keys())
        else:
            keys = s

        if type(k) in complex_types:
            k = k.rep()

        result = self.driver.hmget('{}{}'.format(self.p, k), keys)
        result = [self.decode_value(r) for r in result]

        for i in range(len(keys)):
            d[keys[i]] = result[i]

        return d

    def __getitem__(self, k):
        item = self.get(k)
        if item is None and self.value_type is not None:
            return vivify('{}.{}'.format(self.prefix, k), self.value_type)
        return item

    def __setitem__(self, k, v):
        return self.set(k, v)

    def exists(self, k):
        return self.driver.exists(k)

    def rep(self):
        d = '({'
        for k, v in self.schema.items():
            d += '{}'.format(k)
            d += ':'
            d += encode_type(v)
            d += ','
        d = d[:-1]
        d += '})'
        return CTP + self.rep_str + '<' + self.contract_id + ':' + self.prefix + '>' + d


def table(prefix=None, key_type=str, schema=None):
    if prefix is None:
        return TablePlaceholder(key_type=key_type, schema=schema)
    return Table(prefix=prefix, key_type=key_type, schema=schema)


class Ranked(RObject):
    def __init__(self, prefix=None, key_type=str, value_type=None):
        super().__init__(prefix=prefix,
                         key_type=key_type,
                         value_type=value_type,
                         rep_str='ranked')

    def add(self, member, score: int):
        m = self.encode_value(member, explicit=True)
        return self.driver.zadd(self.prefix, score, m)

    def delete(self, member):
        m = self.encode_value(member, explicit=True)
        return self.driver.zrem(self.prefix, m)

    def get_max(self):
        m = self.driver.zrevrangebyscore(self.prefix, max='+inf', min='-inf', start=0, num=1)
        m = m.pop()
        m = self.decode_value(m)
        return m

    def get_min(self):
        m = self.driver.zrangebyscore(self.prefix, min='-inf', max='+inf', start=0, num=1)
        m = m.pop()
        m = self.decode_value(m)
        return m

    def pop_max(self):
        m = self.get_max()
        self.delete(m)
        return m

    def pop_min(self):
        m = self.get_min()
        self.delete(m)
        return m

    def score(self, member):
        m = self.encode_value(member)
        return self.driver.zscore(self.prefix, m)

    def increment(self, member, i: int):
        m = self.encode_value(member)
        return self.driver.zincrby(self.prefix, m, i)

    def decrement(self, member, i: int):
        i *= -1
        self.increment(member, i)

    def rep(self):
        return '{}ranked<{}:{}>({},{})'.format(CTP,
                                            self.contract_id,
                                            self.prefix,
                                            encode_type(self.key_type),
                                            encode_type(self.value_type))


def ranked(prefix=None, key_type=str, value_type=int):
    if prefix is None:
        return RankedPlaceholder(key_type=key_type, value_type=value_type)
    else:
        return Ranked(prefix=prefix, key_type=key_type, value_type=value_type)


complex_types = [HMap, HList, Table, Ranked]



from seneca.libs.datatypes import hmap

balances = hmap('balances', str, int)
allowed = hmap('allowed', str, hmap(value_type=int))
market = hmap('market', str, int)

@seed
def deposit_to_all_wallets():
    market['stamps_to_tau'] = 1
    balances['lorde'] = 666

@export
def submit_stamps(stamps):
    amount = stamps / market['stamps_to_tau']
    transfer('black_hole', int(amount))

@export
def balance_of(wallet_id):
    return balances[wallet_id]

@export
def transfer(to, amount):
    print("transfering from {} to {} with amount {}".format(rt['sender'], to, amount))
    balances[rt['sender']] -= amount
    balances[to] += amount
    sender_balance = balances[rt['sender']]

    assert sender_balance >= 0, "Sender balance must be non-negative!!!"

@export
def approve(spender, amount):
    allowed[rt['sender']][spender] = amount


@export
def transfer_from(_from, to, amount):
    assert allowed[_from][rt['sender']] >= amount
    assert balances[_from] >= amount

    allowed[_from][rt['sender']] -= amount
    balances[_from] -= amount
    balances[to] += amount


@export
def allowance(approver, spender):
    return allowed[approver][spender]


@export
def mint(to, amount):
    # print("minting {} to wallet {}".format(amount, to))
    assert rt['sender'] == rt['author'], 'Only the original contract author can mint!'

    balances[to] += amount



import base64

exports = {'base64': base64}


def run_tests(_):
    '''
    >>> type(exports)
    <class 'dict'>
    '''
    import doctest, sys
    return doctest.testmod(sys.modules[__name__], extraglobs={**locals()})



from seneca.libs.datatypes import hmap

floats = hmap('floats', str, float)

@export
def store_float(s, f):
    floats[s] = f

@export
def read_float(s):
    return floats[s]

@export
def divide_float(s):
    return floats[s] / 2

@export
def add_floats(s1, s2):
    return floats[s1] + floats[s2]



from seneca.libs.datatypes import *   # This code most be interpretted for the metaclass stuff to get run
from seneca.engine.conflict_resolution import *
from seneca.engine.cr_commands import *
import redis
from unittest import TestCase
import unittest


class TestConflictResolution(TestCase):

    def setUp(self):
        self.master = redis.StrictRedis(host='localhost', port=6379, db=0)
        self.working = redis.StrictRedis(host='localhost', port=6379, db=1)
        self.sbb_data = {}
        self._set_rp()

    def tearDown(self):
        self.master.flushdb()
        self.working.flushdb()

    def _set_rp(self, sbb_idx=0, contract_idx=0, finalize=False):
        if contract_idx in self.sbb_data:
            data = self.sbb_data[contract_idx]
        else:
            data = self._new_cr_data(sbb_idx=sbb_idx, finalize=finalize)
            self.sbb_data[contract_idx] = data

        self.r = RedisProxy(sbb_idx=sbb_idx, contract_idx=contract_idx, data=data)

    def _new_cr_data(self, sbb_idx=0, finalize=False):
        return CRContext(working_db=self.working, master_db=self.master, sbb_idx=sbb_idx, finalize=finalize)

    def test_all_keys_and_values_for_basic_set_get(self):
        KEY1, VAL1 = 'k1', b'v1'
        KEY2, VAL2 = 'k2', b'v2'
        KEY3, VAL3 = 'k3', b'v3'
        NEW_VAL1 = b'v1_NEW'
        NEW_VAL3 = b'v3_NEW'

        # Seed keys on master
        self.master.set(KEY1, VAL1)
        self.master.set(KEY2, VAL2)
        self.working.set(KEY3, VAL3)

        self.r.set(KEY1, NEW_VAL1)
        self.r.contract_idx = 1
        self.r.get(KEY2)  # To trigger a copy to sbb specific layer
        self.r.contract_idx = 2
        self.r.set(KEY3, NEW_VAL3)  # To trigger a copy to sbb specific layer

        # Check the modified and original values
        getset = self.r.data['getset']
        k1_expected = {'og': VAL1, 'mod': NEW_VAL1, 'contracts': {0}}
        k2_expected = {'og': VAL2, 'mod': None, 'contracts': {1}}
        k3_expected = {'og': VAL3, 'mod': NEW_VAL3, 'contracts': {2}}
        self.assertEqual(getset[KEY1], k1_expected)
        self.assertEqual(getset[KEY2], k2_expected)
        self.assertEqual(getset[KEY3], k3_expected)

        # Check modifications list
        expected_mods = {0: {KEY1}, 2: {KEY3}}
        self.assertEqual(self.r.data['getset'].writes, expected_mods)

        # Check should_rerun (tinker with common first)
        cr_data = self.sbb_data[0]['getset']
        self.working.set(KEY1, b'A NEW VALUE HAS ARRIVED')
        self.working.set(KEY2, b'A NEW VALUE HAS ARRIVED AGAIN')
        self.assertTrue(0 in list(cr_data.get_rerun_list(reset_keys=False)))
        self.assertTrue(1 in list(cr_data.get_rerun_list(reset_keys=False)))
        self.assertFalse(2 in list(cr_data.get_rerun_list(reset_keys=False)))

    def test_merge_to_common(self):
        KEY1, VAL1 = 'k1', b'v1'
        KEY2, VAL2 = 'k2', b'v2'
        KEY3, VAL3 = 'k3', b'v3'
        KEY4, VAL4 = 'k4', b'v4'
        NEW_VAL1 = b'v1_NEW'
        NEW_VAL3 = b'v3_NEW'
        NEW_VAL4 = b'v4_NEW'

        # Seed keys on master
        self.master.set(KEY1, VAL1)
        self.master.set(KEY2, VAL2)
        self.master.set(KEY3, b'val 3 on master that should be ignored in presence of KEY3 on common layer')
        self.working.set(KEY3, VAL3)

        self.r.set(KEY1, NEW_VAL1)
        self.r.contract_idx = 2
        self.r.get(KEY2)  # To trigger a copy to sbb specific layer
        self.r.contract_idx = 2
        self.r.set(KEY3, NEW_VAL3)
        self.r.set(KEY4, NEW_VAL4)

        # Check merge_to_common
        cr_data = self.sbb_data[0]
        cr_data.merge_to_common()
        self.assertEqual(self.working.get(KEY1), NEW_VAL1)
        self.assertEqual(self.working.get(KEY2), None)  # None b/c we never SET KEY2
        self.assertEqual(self.working.get(KEY3), NEW_VAL3)
        self.assertEqual(self.working.get(KEY4), NEW_VAL4)

    def test_merge_to_master(self):
        KEY1, VAL1 = 'k1', b'v1'
        KEY2, VAL2 = 'k2', b'v2'
        KEY3, VAL3 = 'k3', b'v3'
        KEY4, VAL4 = 'k4', b'v4'
        NEW_VAL1 = b'v1_NEW'
        NEW_VAL3 = b'v3_NEW'
        NEW_VAL4 = b'v4_NEW'

        # Seed keys on master
        self.master.set(KEY1, VAL1)
        self.master.set(KEY2, VAL2)
        self.master.set(KEY3, b'val 3 on master that should be ignored in presence of KEY3 on common layer')
        self.working.set(KEY3, VAL3)

        self.r.set(KEY1, NEW_VAL1)
        self.r.contract_idx = 2
        self.r.get(KEY2)  # To trigger a copy to sbb specific layer
        self.r.contract_idx = 2
        self.r.set(KEY3, NEW_VAL3)
        self.r.set(KEY4, NEW_VAL4)

        # First merge_to_common
        cr_data = self.sbb_data[0]
        cr_data.merge_to_common()

        # Now check merge_to_master
        CRContext.merge_to_master(working_db=cr_data.working_db, master_db=cr_data.master_db)
        self.assertEqual(self.master.get(KEY1), NEW_VAL1)
        self.assertEqual(self.master.get(KEY2), VAL2)
        self.assertEqual(self.master.get(KEY3), NEW_VAL3)
        self.assertEqual(self.master.get(KEY4), NEW_VAL4)

    def test_state_rep(self):
        KEY1, VAL1 = 'k1', 'v1'
        KEY2, VAL2 = 'k2', 'v2'
        KEY3, VAL3 = 'k3', 'v3'
        KEY4, VAL4 = 'k4', 'v4'
        NEW_VAL1 = 'v1_NEW'
        NEW_VAL3 = 'v3_NEW'
        NEW_VAL4 = 'v4_NEW'

        # Seed keys on master
        self.master.set(KEY1, VAL1)
        self.master.set(KEY2, VAL2)
        self.master.set(KEY3, 'val 3 on master that should be ignored in presence of KEY3 on common layer')
        self.working.set(KEY3, VAL3)

        self.r.set(KEY1, NEW_VAL1)
        self.r.contract_idx = 1
        self.r.get(KEY2)  # To trigger a copy to sbb specific layer
        self.r.contract_idx = 2
        self.r.set(KEY3, NEW_VAL3)
        self.r.set(KEY4, NEW_VAL4)

        # Manually add the contracts/results being run. Store them so we can assert on them later
        expected_contracts = []
        expected_results = []
        for i in range(3):
            contract = 'contract_{}'.format(i)
            result = 'run_result_{}'.format(i)
            self.sbb_data[0].contracts.append(contract)
            self.sbb_data[0].run_results.append(result)
            expected_contracts.append(contract)
            expected_results.append(result)

        # Check entire subblock state
        expected_state = "SET {k1} {v1};SET {k3} {v3};SET {k4} {v4};"\
                         .format(k1=KEY1, v1=NEW_VAL1, k2=KEY2, v2=VAL2, k3=KEY3, v3=NEW_VAL3, k4=KEY4, v4=NEW_VAL4)
        self.assertEqual(self.sbb_data[0]['getset'].get_state_rep(), expected_state)

        # Check individual contract states
        cr_data = self.r.data
        state_0 = "SET {} {};".format(KEY1, NEW_VAL1)
        state_1 = ""
        state_2 = "SET {} {};SET {} {};".format(KEY3, NEW_VAL3, KEY4, NEW_VAL4)
        self.assertEqual(state_0, cr_data['getset'].get_state_for_idx(0))
        self.assertEqual(state_0, cr_data.get_state_for_idx(0))
        self.assertEqual(state_1, cr_data['getset'].get_state_for_idx(1))
        self.assertEqual(state_1, cr_data.get_state_for_idx(1))
        self.assertEqual(state_2, cr_data['getset'].get_state_for_idx(2))
        self.assertEqual(state_2, cr_data.get_state_for_idx(2))

        # Check sb rep...fake merged_to_common
        cr_data.merged_to_common = True
        expected_states = (state_0, state_1, state_2)
        expected_rep = []
        for i in range(3):
            expected_rep.append((expected_contracts[i], expected_results[i], expected_states[i]))

        self.assertEqual(expected_rep, cr_data.get_subblock_rep())

    def test_all_keys_and_values_for_basic_hset_hget(self):
        KEY1, KEY2 = 'KEY1', 'KEY2'
        FIELD1, VAL1 = 'k1', b'v1'
        FIELD2, VAL2 = 'k2', b'v2'
        FIELD3, VAL3 = 'k3', b'v3'
        NEW_VAL1 = b'v1_NEW'
        NEW_VAL3 = b'v3_NEW'

        # Seed keys on master
        self.master.hset(KEY1, FIELD1, VAL1)
        self.master.hset(KEY1, FIELD2, VAL2)
        self.master.hset(KEY2, FIELD3, b'val 3 on master that should be ignored in presence of KEY3 on common layer')
        self.working.hset(KEY2, FIELD3, VAL3)

        self.r.hset(KEY1, FIELD1, NEW_VAL1)
        self.r.contract_idx = 2
        self.r.hget(KEY1, FIELD2)  # To trigger a copy to sbb specific layer
        self.r.contract_idx = 2
        self.r.hset(KEY2, FIELD3, NEW_VAL3)  # To trigger a copy to sbb specific layer

        # Check the modified and original values
        hm = self.r.data['hm']
        k1_expected = {'og': VAL1, 'mod': NEW_VAL1}
        k2_expected = {'og': VAL2, 'mod': None}
        k3_expected = {'og': VAL3, 'mod': NEW_VAL3}
        self.assertEqual(hm[KEY1][FIELD1], k1_expected)
        self.assertEqual(hm[KEY1][FIELD2], k2_expected)
        self.assertEqual(hm[KEY2][FIELD3], k3_expected)

        print("Type of key k1: {}".format(self.master.type(KEY1)))

        # Check modifications list
        # TOFO fix and implement mod list for hmaps
        expected_mods = {KEY1: {FIELD1}, KEY2: {FIELD3}}
        # self.assertEqual(self.r.data['hm'].mods, expected_mods)

        # Check should_rerun (tinker with common first)
        # self.working.hset(FIELD1, b'A NEW VALUE HAS ARRIVED')
        # self.working.hset(FIELD2, b'A NEW VALUE HAS ARRIVED AGAIN')
        # self.assertTrue(self.sbb_data[0].should_rerun(0))
        # self.assertFalse(self.sbb_data[0].should_rerun(1))
        # self.assertFalse(self.sbb_data[0].should_rerun(2))

        # Check merge_to_common
        # cr_data = self.sbb_data[0]
        # cr_data.merge_to_common()
        # self.assertEqual(self.master.get(FIELD1), NEW_VAL1)
        # self.assertEqual(self.master.get(FIELD2), VAL2)
        # self.assertEqual(self.master.get(FIELD3), NEW_VAL3)

        # Check state
        # expected_state = "HNSET {k1} {f1};b'{v1};SET {k2};b'{v2}SET;{k3};b'{v3}"\
        #                  .format(k1=FIELD1, v1=NEW_VAL1, k2=KEY2, v2=VAL2, k3=KEY3, v3=VAL3)

    def test_unimplemented_method_raises_assert(self):
        with self.assertRaises(AssertionError):
            self.r.this_is_not_implemented('some_key')


if __name__ == "__main__":
    unittest.main()




from seneca.libs.datatypes import hmap


balances = hmap('balances', str, int)


@export
def test_global_namespace():
    print('sender: {}, author: {}'.format(rt['sender'], rt['author']))
    print("sbb_idx: {}".format(sbb_idx))
    print("ALL GLOBALS: {}".format(globals()))



balances = {'hello': 'world'}

@export
def one_you_can_export():
    print('Running one_you_can_export()')

@export
def one_you_can_also_export():
    print('Running one_you_can_also_export()')
    one_you_can_export()

def one_you_cannot_export(dont, do, it='wrong'):
    print('Always runs: Running one_you_cannot_export()')

@export
def one_you_can_also_also_export():
    print('Running one_you_can_also_also_export()')
    one_you_cannot_export('a', 'b', it='c')



import redis
from collections import defaultdict
from seneca.libs.logger import get_logger
from typing import List
# TODO -- clean this file up


class RedisOperation:
    def __init__(self, op_name: str, key: str, *args, **kwargs):
        self.op_name, self.key, self.args, self.kwargs = op_name, key, args, kwargs


class CRDataMeta(type):
    def __new__(cls, clsname, bases, clsdict):
        clsobj = super().__new__(cls, clsname, bases, clsdict)
        if not hasattr(clsobj, 'registry'):
            clsobj.registry = {}

        # Only add classes that have the 'NAME' field set
        if 'NAME' in clsdict:
            clsobj.registry[clsdict['NAME']] = clsobj
        return clsobj


class CRDataBase(metaclass=CRDataMeta):
    def __init__(self, master_db: redis.StrictRedis, working_db: redis.StrictRedis):
        super().__init__()
        self.log = get_logger(type(self).__name__)
        self.master, self.working = master_db, working_db
        self.writes = defaultdict(set)
        self.reads = defaultdict(set)
        self.outputs = defaultdict(str)
        self.redo_log = defaultdict(dict)

    def merge_to_common(self):
        """ Merges the subblock specific data to the common layer """
        raise NotImplementedError()

    def get_state_rep(self) -> str:
        """ Updates the 'state' list for the changes represented in this data structure. The state list is a list of outputs
        or modifications from every contract. """
        raise NotImplementedError()

    def get_state_for_idx(self, contract_idx: int) -> str:
        """
        Gets a state representation string for a particular contract index. This should be overwritten by all subclasses
        that track any sort of state modifications """
        return ''

    def reset_contract_data(self, contract_idx: int):
        """ Resets the reads list and modification list for the contract at index idx. """
        self.writes[contract_idx].clear()
        self.reads[contract_idx].clear()
        self.outputs[contract_idx] = ''

    # TODO better tooling
    # Abstraction for get_modified_keys/reset_keys is very weak. I don't think they will work with complex data types
    def get_modified_keys(self) -> set:
        return set()

    def get_modified_keys_recursive(self) -> set:
        return set()

    def reset_key(self, key):
        pass

    def get_rerun_list(self, reset_keys=True) -> List[int]:
        return []

    def rollback_contract(self, contract_idx: int):
        pass


class CRDataGetSet(CRDataBase, dict):
    NAME = 'getset'

    def _get_modified_keys(self):
        # TODO this needs to return READs that have had their original values changed too!
        return set().union((key for key in self if self[key]['og'] != self[key]['mod'] and self[key]['mod'] is not None))

    def merge_to_common(self):
        modified_keys = self._get_modified_keys()
        for key in modified_keys:
            self.working.set(key, self[key]['mod'])

    @classmethod
    def merge_to_master(cls, working_db: redis.StrictRedis, master_db: redis.StrictRedis, key: str):
        assert working_db.exists(key), "Key {} must exist in working_db to merge to master".format(key)
        val = working_db.get(key)
        master_db.set(key, val)

    def get_state_rep(self) -> str:
        """
        Return a representation of all redis DB commands to update to the absolute state in minimum operations
        :return: A string with all redis command in raw executable form, delimited by semicolons
        """
        modified_keys = self._get_modified_keys()
        # Need to sort the modified_keys so state output is deterministic
        return ''.join('SET {} {};'.format(k, self[k]['mod'].decode()) for k in sorted(modified_keys))

    def rollback_contract(self, contract_idx: int):
        self.log.debug("Reseting contract idx {}".format(contract_idx))
        if contract_idx not in self.redo_log:
            # TODO for dev, we raise an exception, as we do not expect contracts to read only w/o writing
            # raise Exception("Contract idx {} not in redo_log!".format(contract_idx))
            self.log.warning("Contract idx {} not in redo_log! Returning without any reverts".format(contract_idx))
            return

        self.reset_contract_data(contract_idx)

        for key in self.redo_log[contract_idx]:
            og_val = self.redo_log[contract_idx][key]
            # Remove the key entirely if value is none
            if og_val is None:
                self.log.debugv("Removing key {}".format(key))
                del self[key]
            # Otherwise, reset the key to the value before the contract
            else:
                self.log.debugv("Resetting key {} to value {}".format(key, og_val))
                self[key]['mod'] = og_val

            # Remove this contract idx from the key's affected contracts
            if contract_idx in self[key]['contracts']:
                self[key]['contracts'].remove(contract_idx)

    def get_state_for_idx(self, contract_idx: int) -> str:
        return self.outputs[contract_idx]

    def revert_contract(self, contract_idx: int):
        assert contract_idx in self.redo_log, "Contract index {} not found in redo log!".format(contract_idx)

    def get_rerun_list(self, reset_keys=True) -> List[int]:
        mod_keys = self.get_modified_keys_recursive()
        contract_set = set()
        self.log.debugv("Modified keys for rerunning: {}".format(mod_keys))

        for key in mod_keys:
            contract_set = contract_set.union(self[key]['contracts'])
            if reset_keys:
                self.reset_key(key)

        self.log.debugv("CONTRACT SET TO RERUN: {}".format(contract_set))

        return sorted(contract_set)

    def get_modified_keys(self) -> set:
        mods = set()
        for k in self:
            if (self.master.exists(k) and (self.master.get(k) != self[k]['og'])) or (
                    self.working.exists(k) and (self.working.get(k) != self[k]['og'])):
                mods.add(k)

        return mods

    def get_modified_keys_recursive(self) -> set:
        mod_keys = self.get_modified_keys()
        self.add_adjacent_keys(mod_keys)
        return mod_keys

    def add_adjacent_keys(self, key_set):
        copy_set = set(key_set)
        for key in copy_set:
            self._add_adjacent_keys(key, key_set)

    def _add_adjacent_keys(self, key: str, key_set: set):
        assert key in key_set, 'logic error'
        assert key in self, 'key is not in self??'

        # Get all keys modified in conjunction with 'key'
        new_keys = set()
        for contract_idx in self[key]['contracts']:
            all_rw = self.writes[contract_idx].union(self.reads[contract_idx])
            new_keys = new_keys.union(all_rw)

        # Recursive stage
        for k in new_keys:
            if k in key_set:  # Base case -- if this key is already in the key_list do not recurse
                continue
            else:
                key_set.add(k)
                self._add_adjacent_keys(k, key_set)

    def reset_key(self, key):
        # TODO I  think we are going to have problems when the key is different on both master AND common.
        # this is because if this context was built on top of another pending_db, and we are not the first subblock,
        # should we use the common layer (which presumably a prior sb copied from the updated master), or do we use
        # the updated master? I think we need some special logic for if this is subblock 0 then use master, otherwise
        # use common
        self.log.debugv("Resetting key {}".format(key))
        og_val = self[key]['og']

        self[key]['mod'] = None
        self[key]['contracts'] = set()

        # First, try and copy over master if it differs from original value

        if self.working.exists(key) and self.working.get(key) != og_val:
            self.log.debugv("Reseting key {} to COMMON value {}".format(key, self.working.get(key)))
            self[key]['og'] = self.working.get(key)
        elif self.master.exists(key) and self.master.get(key) != og_val:
            self.log.debugv("Reseting key {} to MASTER value {}".format(key, self.master.get(key)))
            self[key]['og'] = self.master.get(key)
        else:
            self.log.spam("No updated value found for key {}. Clearing modified and leaving original val".format(key))


class CRDataHMap(CRDataBase, defaultdict):
    NAME = 'hm'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_factory = dict

    def _get_modified_keys(self) -> dict:
        """
        Returns a dict of sets. Key is key in the hmap, and set is a list of modified fields for that key
        """
        mods_dict = defaultdict(set)
        for key in self:
            for field in self[key]:
                if self[key][field]['og'] != self[key][field]['mod']:
                    mods_dict[key].add(field)

        return mods_dict

    def merge_to_common(self):
        return False  # TODO implement
        raise NotImplementedError()

    def get_state_rep(self):
        return False  # TODO implement
        raise NotImplementedError()

    @classmethod
    def merge_to_master(cls, working_db: redis.StrictRedis, master_db: redis.StrictRedis, key: str):
        assert working_db.exists(key), "Key {} must exist in working_db to merge to master".format(key)

        all_fields = working_db.hkeys(key)
        for field in all_fields:
            val = working_db.hget(key, field)
            master_db.hset(key, field, val)


class CRDataDelete(CRDataBase, set):
    NAME = 'del'

    def merge_to_common(self):
        return False  # TODO implement
        raise NotImplementedError()

    def get_state_rep(self):
        return False  # TODO implement
        raise NotImplementedError()


class CRDataOperations(CRDataBase, list):
    """
    CRDataOperations is a list of RedisOperation instances.
    """
    NAME = 'ops'

    def merge_to_common(self):
        pass  # TODO implement
        # raise NotImplementedError()

    def get_state_rep(self):
        pass  # TODO implement
        # raise NotImplementedError()

    def should_rerun(self, contract_idx: int) -> bool:
        pass  # TODO implement
        # raise NotImplementedError()


class CRDataOutputs(CRDataBase, list):
    """
    This structure is a list of tuples. The index of the outer list correspons to the output of the contract with that
    same index. The tuple itself always has 2 elements, and is of the form [RESULT, OUTPUT], where
    """
    NAME = 'out'

    def merge_to_common(self):
        return False  # TODO implement
        raise NotImplementedError()

    def get_state_rep(self):
        return False  # TODO implement
        raise NotImplementedError()


class CRContext:

    def __init__(self, working_db: redis.StrictRedis, master_db: redis.StrictRedis, sbb_idx: int, finalize=False):
        self.log = get_logger(type(self).__name__)
        # TODO do all these fellas need to be passed in? Can we just grab it from the Bookkeeper? --davis
        self.finalize = finalize
        self.working_db, self.master_db = working_db, master_db
        self.sbb_idx = sbb_idx

        # cr_data holds instances of CRDataBase. The key is the 'NAME' field specified in the CRDataBase subclass
        # For convenience, all these keys are directly accessible from this CRContext instance (see __getitem__)
        self.cr_data = {name: obj(master_db=self.master_db, working_db=self.working_db) for name, obj in
                        CRDataBase.registry.items()}

        # TODO deques are probobly more optimal than using arrays here
        # run_results is a list of strings, representing the return code of contracts (ie 'SUCC', 'FAIL', ..)
        self.run_results = []
        self.contracts = []  # A list of ContractionTransaction objects. SenecaClient should append as it runs contracts
        self.input_hash = None  # Input hash should be set by SenecaClient once a new sub block is started
        self.merged_to_common = False

    @property
    def next_contract_idx(self):
        assert len(self.contracts) == len(self.run_results), "Oh dear...a logic error is present"  # TODO remove
        return len(self.contracts)

    def add_contract_result(self, contract, result: str):
        assert len(self.contracts) == len(self.run_results), "Oh dear...a logic error is present"  # TODO remove
        self.contracts.append(contract)
        self.run_results.append(result)

    def update_contract_result(self, contract_idx: int, result: str):
        assert len(self.contracts) == len(self.run_results), "Oh dear...a logic error is present"  # TODO remove
        assert len(self.contracts) > contract_idx, "contract_idx {} out of bounds. Only {} contracts in self.contracts"\
                                                   .format(contract_idx, len(self.contracts))
        self.log.debugv("Updating run result for contract idx {} to <{}>".format(contract_idx, result))
        self.run_results[contract_idx] = result

    def rollback_contract(self, contract_idx: int):
        # TODO this only works for set/get
        self.cr_data['getset'].rollback_contract(contract_idx)

    def reset(self, reset_db=True):
        """ Resets all state held by this container. """
        # TODO i think this would be a lot easier if we just scrapped this whole CRContext object and made a new
        # one, but then would we have to worry about memory leaks? idk but either way screw python
        def _is_subclass(obj, subs: tuple):
            """ Utility method. Returns true if 'obj' is a subclass of any of the classes in subs """
            for s in subs:
                if issubclass(type(obj), s): return True
            return False

        self.log.debug("Reseting CRData with reset_db={}".format(reset_db))
        if reset_db:
            self.working_db.flushdb()

        # Reset this object's state
        self.merged_to_common = False
        self.input_hash = None
        self.run_results.clear()
        self.contracts.clear()
        # TODO is this ok resetting all the CRData's like this? Should we worry about memory leaks? --davis
        self.cr_data = {name: obj(master_db=self.master_db, working_db=self.working_db) for name, obj in
                        CRDataBase.registry.items()}

    def get_state_for_idx(self, contract_idx: int) -> str:
        """
        Returns the state for the contract at the specified index
        """
        assert contract_idx < len(self.contracts), "Contract index {} out of bounds for self.contracts of length {}" \
                                                   .format(contract_idx, len(self.contracts))

        state_str = ''
        for key in sorted(self.cr_data.keys()):  # We sort the keys so that output will always be deterministic
            state_str += self.cr_data[key].get_state_for_idx(contract_idx)
        return state_str

    def get_subblock_rep(self) -> List[tuple]:
        """
        Returns a list of tuples. There will be one tuple for each contract in self.contracts, and tuples will be of the
        form (contract, status, state). contract will be an instance of ContractTransaction. Status will be a string
        representing the execution status of the contract (fail/succ/ect). State will be a string that represents the
        changes to state made by that contract.
        """
        assert len(self.contracts) == len(self.run_results), "you done shit the bed again davis"
        assert self.merged_to_common, "You should have merged to common before trying to get the subblock rep"

        return [(self.contracts[i], self.run_results[i], self.get_state_for_idx(i)) for i in range(len(self.contracts))]

    def iter_rerun_indexes(self):
        # TODO this only works for getset right now
        # TODO this does not support new keys being modified during the rerun process
        data = self.cr_data['getset']
        contract_list = data.get_rerun_list()
        self.log.debugv("Contracts indexes to rerun: {}".format(contract_list))

        for i in contract_list:
            self.log.debugv("Rerunning contract at index {}".format(i))
            og_reads, og_writes = data.reads[i], data.reads[i]
            self.reset_contract_data(i)

            yield i

            # TODO i think these assertions will fail when we rerun a contract that fails (b/c read/write list will
            # be empty). We should only do this logic if the execution was successful
            assert og_reads == data.reads[i], "Original reads have changed for contract idx {}!\nOriginal: {}\nNew " \
                                              "Reads: {}".format(i, og_reads, data.reads[i])
            assert og_writes == data.writes[i], "Original writes have changed for contract idx {}!\nOriginal: {}\nNew " \
                                                "Writes: {}".format(i, og_writes, data.writes[i])

    def reset_contract_data(self, contract_idx: int):
        """
        Resets the reads list and modification list for the contract at index idx.
        """
        for obj in self.cr_data.values():
            obj.reset_contract_data(contract_idx)

    def merge_to_common(self):
        assert not self.merged_to_common, "Already merged to common! merge_to_common should only be called once"

        for obj in self.cr_data.values():
            obj.merge_to_common()

        self.merged_to_common = True

    @classmethod
    def merge_to_master(cls, working_db: redis.StrictRedis, master_db: redis.StrictRedis):
        from seneca.engine.client import Macros  # to avoid cyclic imports

        for key in working_db.keys():
            # Ignore Phase keys
            if key in Macros.ALL_MACROS:
                continue

            t = working_db.type(key)

            if t == b'string':
                CRDataGetSet.merge_to_master(working_db, master_db, key)
            elif t == b'hash':
                CRDataHMap.merge_to_master(working_db, master_db, key)
            else:
                raise NotImplementedError("No logic implemented for copying key <{}> of type <{}>".format(key, t))

    def __getitem__(self, item):
        assert item in self.cr_data, "No structure named {} in cr_data. Only keys available: {}"\
                                     .format(item, list(self.cr_data.keys()))
        return self.cr_data[item]


class RedisProxy:

    def __init__(self, sbb_idx: int, contract_idx: int, data: CRContext, finalize=False):
        # TODO do all these fellas need to be passed in? Can we just grab it from the Bookkeeper? --davis
        self.finalize = finalize
        self.data = data
        self.working_db, self.master_db = data.working_db, data.master_db
        self.sbb_idx, self.contract_idx = sbb_idx, contract_idx

    def __getattr__(self, item):
        from seneca.engine.cr_commands import CRCmdBase  # To avoid cyclic imports -- TODO better solution?
        assert item in CRCmdBase.registry, "redis operation {} not implemented for conflict resolution".format(item)

        return CRCmdBase.registry[item](working_db=self.working_db, master_db=self.master_db,
                                        sbb_idx=self.sbb_idx, contract_idx=self.contract_idx, data=self.data,
                                        finalize=self.finalize)


# print("CRDataMetaRegistery")
# for k, v in CRDataBase.registry.items():
#     print("{}: {}".format(k, v))


"""
THOUGHT

each CRData has state variables for this rerun process, including
- internal list of contracts that HAVE been rerun?
- 

SHOULD WE ALSO MAINTAIN a mapping of keys to contracts that read/write them? Otherwise we have to do this o(n) 
everytime.. 

JUST RAISE AN ASSERTION FOR NOW IF A NEW KEY IS MODIFIED 

1) at start of rerun, CRData gets all contracts that have had their original values changed (on common or master)
2) copy over new values into effected key's original values (prioritize master if master is diff). Set mod value to None
3) if 

       # build a set of all reads/write that have their original value changed
        # copy, from common layer, to the new original value, and set the modified value to None
        # build a min heap of contract indexes that need to be run by check contract's mod list
        # reset the contract data before you rerun it
        # loop

        # PREFER MASTER VALUE when copying keys over. if og should have been copied from master during exec phase,
        # so if master is diff from og that means another block changed master, and that value should be prefered

        # what if both master and common differ from the orig values? which ones do you use?
        # we would need to track original master value (at time of the read). or just ignore this problem until
        # we implement proper chaining of db reads

"""



from seneca.libs.datatypes import hmap, ranked, hlist

now = 12345 # unix timestamp


class Context:
    def __init__(self, sender=None, author=None):
        self.sender = sender
        self.author = author


rt = Context()

current_delegates = hlist('delegates', str)
votes = ranked('votes')
last_vote = hmap('last_vote', str, int)

defaults = hmap('defaults', str, None)
defaults.set('voting_period', 10000)
defaults.set('term', 100000)
defaults.set('last_election', 0)
defaults.get('in_vote', False)
defaults.get('delegate_count', 63)

def vote_for_delegate(d):
    last_election = defaults.get('last_election')
    if last_vote.get(rt['sender']) != last_election:
        last_vote.set(last_election)
        return votes.increment(d, 1)


def initiate_vote():
    # make sure we are not in a vote
    if not defaults.get('in_vote'):
        last_election = defaults.get('last_election')
        term = defaults.get('term')
        voting_period = defaults.get('voting_period')

        # make sure that enough time has passed to start a new vote
        if now > last_election + term:
            defaults.set('in_vote', True)

            # set the last election to a point in the future so that we can easily see if its time to end the vote
            defaults.set('last_election', now + voting_period)
            votes.drop()


def end_vote():
    if defaults.get('in_vote'):
        last_election = defaults.get('last_election')

        if now > last_election:
            defaults.set('in_vote', False)
            count = defaults.get('delegate_count')
            top_delegates = map(votes.pop_max(), range(count))
            current_delegates.drop()
            map(lambda x: current_delegates.append(x), top_delegates)



from setuptools import setup, find_packages
from setuptools.extension import Extension
from distutils.command.build_ext import build_ext
import os

major = 0

def get_version_number():
    if os.getenv('CIRCLECI'):
        minor, patch = divmod(int(os.getenv('CIRCLE_BUILD_NUM')), 180)
        ver = '{}.{}.{}'.format(major, minor, patch)
        with open('seneca/.version', 'w+') as f:
            f.write(ver)
        return ver
    else:
        with open('seneca/.version') as f:
            ver = f.read()
            return ver

__version__ = get_version_number()

print('#' * 128)
print('\n    VERSION = {}\n'.format(__version__))
print('#' * 128)

requirements = [
    'redis==2.10.6',
    'python-dotenv==0.9.1',
    'ujson==1.35',
    'autopep8==1.4.3'
]

class ve_build_ext(build_ext):
    """Build C extensions, but fail with a straightforward exception."""

    def run(self):
        """Wrap `run` with `BuildFailed`."""
        try:
            build_ext.run(self)
        except errors.DistutilsPlatformError:
            raise BuildFailed()

    def build_extension(self, ext):
        """Wrap `build_extension` with `BuildFailed`."""
        try:
            # Uncomment to test compile failure handling:
            #   raise errors.CCompilerError("OOPS")
            build_ext.build_extension(self, ext)
        except ext_errors:
            raise BuildFailed()
        except ValueError as err:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(err):    # works with both py 2/3
                raise BuildFailed()
            raise

setup(
    name='seneca',
    version=__version__,
    description='Python-based smart contract language and interpreter.',
    packages=find_packages(),
    install_requires=requirements,
    url='https://github.com/Lamden/seneca',
    author='Lamden',
    author_email='team@lamden.io',
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    zip_safe=True,
    include_package_data=True,
    ext_modules=[
        Extension('seneca.libs.metering.tracer', sources = ['seneca/libs/metering/tracer.c'])
    ],
    cmdclass={
        'build_ext': ve_build_ext,
    },
)



import hashlib
from functools import partial

'''
    Highly abstracted hashing library that exposes the algorithms guaranteed by hashlib a la Python 3.
    These functions are the following: blake2b, sha3_384, sha3_224, md5, sha256, sha512, sha1, sha224, 
    sha3_256, sha3_512, sha384

    Shake functions aren't available because they are obscure and add complexity to hashing which should be as easy
    as possible to tooling with.
'''

supported_hashing_functions = hashlib.algorithms_guaranteed
supported_hashing_functions.remove('shake_128')
supported_hashing_functions.remove('shake_256')
supported_hashing_functions.remove('blake2s')


def hash_data(d: bytes, algo: str, as_hex=False):
    assert algo in supported_hashing_functions
    assert type(d) == bytes, 'd must be of type bytes.'

    m = hashlib.new(algo)
    m.update(d)
    return m.hexdigest() if as_hex else m.digest()


def f(algo: str):
    assert algo in supported_hashing_functions

    return partial(hash_data, algo=algo)


exports = {
    key: f(key) for key in supported_hashing_functions
}


# explicit tests of the hash functions for sanity
def run_tests(_):
    import doctest, sys

    from seneca.execute import Empty

    res = Empty()
    res.attempted = 0
    res.failed = 0

    payload = b'testing this'
    algos = supported_hashing_functions

    for algo in algos:
        m = hashlib.new(algo)
        m.update(payload)

        m2 = exports[algo](payload)

        try:
            res.attempted += 1
            assert (m.digest() == m2)
        except Exception as e:
            res.failed += 1
    return res


# from clove import constants
# # from clove.network.base import BaseNetwork
# # from clove.utils.external_source import get_transaction
# # networks = constants.NETWORKS_WITH_API
# #
# # print(networks)
# #
# # def audit_contract(network, txid=None):
# #     print(get_transaction(network=network, tx_hash=txid))
# #
# #
# # audit_contract('btc', '5bce375f07f6f373d18ec343b16375fd2accac3b7bde57a1c99f64f93cad8538')


from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter, ReadOnlyException, CompilationException
from os.path import join
from tests.utils import captured_output, TestInterface
import redis, unittest, seneca

test_contracts_path = seneca.__path__[0] + '/../test_contracts/'


class TestScope(TestInterface):

    def test_scope(self):
        """
            Importing exported functions should pass
        """
        self.si.execute_code_str("""
from test_contracts.good import one_you_can_export, one_you_can_also_export, one_you_can_also_also_export
one_you_can_export()
one_you_can_also_export()
one_you_can_also_also_export()
        """)

    def test_scope_fail(self):
        """
            Importing protected functions should fail
        """
        with self.assertRaises(CompilationException) as context:
            self.si.execute_code_str("""
from test_contracts.good import one_you_cannot_export
            """)

    def test_globals(self):
        scope = {'rt': {'sender':'123'}}
        self.si.execute_code_str("""
from test_contracts.reasonable import reasonable_call
result = reasonable_call()
        """, scope)
        self.assertEqual(scope.get('result'), 'sender: 123, contract: test_contracts.reasonable')

    def test_globals_redis(self):
        bk_info = {'sbb_idx': 2, 'contract_idx': 12}
        rt_info = {'rt': {'sender': 'davis', 'author': 'davis'}}
        all_info = {**bk_info, **rt_info}
        with open(join(test_contracts_path, 'sample.sen.py')) as f:
            self.si.publish_code_str('sample', 'davis', f.read())
            self.si.execute_code_str("""
from seneca.contracts.sample import do_that_thing
result = do_that_thing()
            """, all_info)
            self.assertEqual(all_info.get('result'), 'sender: davis, author: davis')

    def test_execute_function(self):
        contracts = ['currency', 'reasonable']
        for contract in contracts:
            with open('{}/{}.sen.py'.format(test_contracts_path, contract)) as f:
                self.si.publish_code_str(contract, 'anonymoose', f.read())

        self.si.execute_function('seneca.contracts.currency.mint',
            'anonymoose', stamps=None, to='anonymoose', amount=10000)

        result = self.si.execute_function('seneca.contracts.reasonable.call_with_args',
            'anonymoose', 10000, 'it is required', not_required='it is not requried')

        self.assertEqual(result['status'], 'success')

    def test_execute_function_invalid(self):
        with open('{}/currency.sen.py'.format(test_contracts_path)) as f:
            self.si.publish_code_str('currency', 'anonymoose', f.read())
        with self.assertRaises(ImportError) as context:
            result = self.si.execute_function('seneca.engine.util.make_n_tup',
                'also_me', 10000, {'x': 'y'})
            print('Should not print this: ', result)

    def test_execute_function_out_of_gas(self):
        contracts = ['currency', 'reasonable']
        for contract in contracts:
            with open('{}/{}.sen.py'.format(test_contracts_path, contract)) as f:
                self.si.publish_code_str(contract, 'anonymoose', f.read())
        self.si.execute_function('seneca.contracts.currency.mint',
            'anonymoose', stamps=None, to='anonymoose', amount=10000)
        with self.assertRaises(AssertionError) as context:
            result = self.si.execute_function('seneca.contracts.reasonable.call_with_args',
                'anonymoose', 5, 'it is required', not_required='it is not requried')

if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
from seneca.engine.interface import SenecaInterface
from seneca.engine.interpreter import SenecaInterpreter, ReadOnlyException, CompilationException
from os.path import join
from tests.utils import captured_output, TestInterface
import redis, unittest, seneca

test_contracts_path = seneca.__path__[0] + '/test_contracts/'

class TestRedis(TestInterface):

    def test_read_only_variables(self):
        with self.assertRaises(CompilationException) as context:
            self.si.execute_code_str("""
__contract__ = 'hacks'
            """)

    def test_read_only_variables_custom(self):
        with self.assertRaises(ReadOnlyException) as context:
            self.si.execute_code_str("""
bird = 'hacks'
            """, {'bird': '123'})

    def test_read_only_variables_aug_assign(self):
        with self.assertRaises(ReadOnlyException) as context:
            self.si.execute_code_str("""
bird += 1
            """, {'bird': 123})

    def test_import_datatypes(self):
        self.si.execute_code_str("""
from seneca.libs.datatypes import hmap
hmap('balance', str, int)
        """)

    def test_import_datatypes_reassign(self):
        with self.assertRaises(ReadOnlyException) as context:
            self.si.execute_code_str("""
from seneca.libs.datatypes import hmap
hmap = 'hacked'
            """)

if __name__ == '__main__':
    unittest.main()



import unittest

from unittest import TestCase
from seneca.engine.interpreter import SenecaInterpreter, Seneca
from seneca.libs.datatypes import *
from seneca.libs.decimal import make_decimal
from seneca.constants.config import get_redis_port, MASTER_DB, DB_OFFSET, get_redis_password

'''
test complex key types and value types
test more failure cases
'''


class TestDatatypes(TestCase):

    def setUp(self):
        Seneca.concurrent_mode = False
        Seneca.loaded = {'__main__': {'rt': {'author': 'me', 'sender': 'me', 'contract': 'seneca.contracts.currency'}}}
        self.r = redis.StrictRedis(host='localhost', port=get_redis_port(), db=MASTER_DB, password=get_redis_password())
        self.l = HList(prefix='yo')
        self.r.flushdb()

    def test_type_mappings(self):
        self.assertTrue(type_to_string[str] == 'str')
        self.assertTrue(type_to_string[int] == 'int')
        self.assertTrue(type_to_string[bool] == 'bool')

        self.assertTrue(string_to_type['str'] == str)
        self.assertTrue(string_to_type['int'] == int)
        self.assertTrue(string_to_type['bool'] == bool)

    def test_parse_representation_map(self):
        repr_str = '*map<seneca.contracts.currency:test>(int,str)'
        m = parse_representation(repr_str)

        self.assertTrue(type(m) == HMap)
        self.assertTrue(m.key_type == int)
        self.assertTrue(m.value_type == str)

    def test_placeholder(self):
        p = Placeholder(placeholder_type=HMap)

        self.assertTrue(p.key_type == str)
        self.assertTrue(p.value_type == int)
        self.assertTrue(p.placeholder_type == HMap)

        good_repr_str = '*map<seneca.contracts.currency:some_map>(str,int)'
        good_map = parse_representation(good_repr_str)

        self.assertTrue(p.valid(good_map))

        bad_repr_str = '*map<seneca.contracts.currency:some_other_map>(int,str)'
        bad_map = parse_representation(bad_repr_str)

        self.assertFalse(p.valid(bad_map))

    def test_robject_encoding_variable(self):
        r = RObject()

        with self.assertRaises(AssertionError):
            r.encode_value('s')

        v = r.encode_value(1)

        self.assertTrue(type(v), bytes)
        self.assertTrue(v.decode(), '1')

    def test_robject_encoding_placeholder(self):
        p = Placeholder(placeholder_type=HMap)
        r = RObject(value_type=p)

        repr_str = '*map<seneca.contracts.currency:howdy>(str,int)'
        _map = parse_representation(repr_str)

        v = r.encode_value(_map)

        self.assertTrue(type(v), bytes)
        self.assertTrue(v.decode(), repr_str)

    def test_robject_decoding_variable(self):
        r = RObject()
        self.assertTrue(r.decode_value(b'1'), 1)
        self.assertTrue(r.decode_value(b'"s"'), 's')
        self.assertTrue(r.decode_value(b'[1, 2, 3]'), [1, 2, 3])

        repr_str = b'*map<seneca.contracts.currency:howdy>(str,int)'
        decoded_map = r.decode_value(repr_str)

        self.assertTrue(type(decoded_map), HMap)
        self.assertTrue(decoded_map.key_type, str)
        self.assertTrue(decoded_map.value_type, int)
        self.assertTrue(decoded_map.prefix, 'howdy')

    def test_robject_check_key_type(self):
        r = RObject()
        with self.assertRaises(AssertionError):
            r.check_key_type(1)

        r.check_key_type('a')

    def test_hmap_set_get(self):
        m = HMap(prefix='howdy')
        m.set('stu', 100)

        s = m.get('stu')

        self.assertEqual(s, 100)

        with self.assertRaises(AssertionError):
            m.set('stu', 'stu')

        n = m.get('not_stu')
        self.assertEqual(n, None)

    def test_hmap_getitem_setitem(self):
        m = HMap(prefix='howdy2')
        m['stu'] = 1000000
        self.assertEqual(m['stu'], 1000000)

        with self.assertRaises(AssertionError):
            m['stu'] = 'stu'

        n = m['not_stu']
        self.assertEqual(n, 0)

    def test_hmap_convenience_function(self):
        m = hmap()
        self.assertTrue(isinstance(m, Placeholder))

        m2 = hmap('howdy')
        self.assertTrue(isinstance(m2, HMap))

    def test_hlist_init_repr(self):
        self.assertEqual(self.l.rep(), '*list<seneca.contracts.currency:yo>(int)')
        self.assertEqual(self.l.prefix, 'yo')

    def test_hlist_push_pop(self):
        self.l.push(123)

        with self.assertRaises(AssertionError):
            self.l.push('123')

        x = self.l.pop()

        self.assertEqual(x, 123)

    def test_hlist_get(self):
        self.l.push(123)
        self.l.push(567)

        x = self.l.get(1)
        y = self.l.get(0)

        self.assertEqual(x, 123)
        self.assertEqual(y, 567)

        # empty the list so we can use it again
        self.l.pop()
        self.l.pop()

    def test_hlist_get_index_error(self):
        self.assertEqual(self.l.get(10000), None)

    def test_hlist_set(self):
        self.l.push(123)
        self.l.push(567)

        self.l.set(1, 890)
        self.l.set(0, 166)

        x = self.l.get(1)
        y = self.l.get(0)

        self.assertEqual(x, 890)
        self.assertEqual(y, 166)

        # empty the list so we can use it again
        self.l.pop()
        self.l.pop()

    def test_hlist_set_index_error(self):

        with self.assertRaises(redis.exceptions.ResponseError):
            self.l.set(10000, 123)

    def test_hlist_pop_right(self):

        self.l.push(123)
        self.l.push(456)

        x = self.l.pop_right()

        self.assertEqual(x, 123)

        self.l.pop()

    def test_hlist_append(self):

        self.l.push(123)
        self.l.append(456)

        x = self.l.pop_right()

        self.assertEqual(x, 456)

        self.l.pop()

    def test_hlist_extend(self):

        self.l.extend([1, 2, 3, 4, 5])

        a = self.l.pop()
        b = self.l.pop()
        c = self.l.pop()
        d = self.l.pop()
        e = self.l.pop()

        self.assertEqual(a, 1)
        self.assertEqual(b, 2)
        self.assertEqual(c, 3)
        self.assertEqual(d, 4)
        self.assertEqual(e, 5)

    def test_hlist_getitem_setitem(self):

        self.l.push(123)
        self.l[0] = 123456789

        self.assertEqual(self.l[0], 123456789)

        self.l.pop()

    def test_hlist_convenience_methods(self):
        p = hlist()

        self.assertTrue(isinstance(p, Placeholder))
        self.assertTrue(p.value_type == int)

        ll = hlist('hello_there')
        self.assertTrue(isinstance(ll, HList))
        self.assertEqual(ll.prefix, 'hello_there')
        self.assertEqual(p.value_type, int)

    def test_hlist_store_placeholders(self):
        complex_l = hlist('goodtimes', value_type=hmap())

        complex_l.push(hmap('some map'))

        m = complex_l.pop()
        self.assertTrue(m.prefix, 'some map')

    def test_table_init(self):
        t = Table(prefix='holla', schema={'name': str, 'balance': int})

        self.assertEqual(t.schema, {'name': str, 'balance': int})
        self.assertEqual(t.p, 'holla:')

        with self.assertRaises(AssertionError):
            bad_t = Table(prefix='bad_boy', schema={'howdy': 'partner'})

        with self.assertRaises(AssertionError):
            bad_t = Table(prefix='oh_no', schema={1: True})

    def test_table_schema_matching(self):
        t = Table(prefix='holla', schema={'name': str, 'balance': int})

        good_dict = {'name': 'stu', 'balance': 100}
        subset_dict = {'name': 'stu'}

        mismatched_type_dict = {'name': 1, 'balance': '100'}
        wrong_keys_dict = {'pizza_hut': 'taco_bell', 'im_at_the': 123}

        t.dict_matches_schema(good_dict)
        t.dict_matches_schema(subset_dict)

        with self.assertRaises(AssertionError):
            t.dict_matches_schema(mismatched_type_dict)

        with self.assertRaises(AssertionError):
            t.dict_matches_schema(wrong_keys_dict)

    def test_table_encoding_variable(self):
        t = Table(prefix='holla', schema={'name': str, 'balance': int})

        v = t.encode_value(1, int)
        self.assertEqual(v, b'1')

        with self.assertRaises(AssertionError):
            t.encode_value('b', int)

    def test_table_encoding_placeholders(self):
        p = ListPlaceholder()

        t = Table(prefix='complex', schema={'name': str, 'list': p})

        repr_str = '*list<seneca.contracts.currency:some_list>(int)'
        l = parse_representation(repr_str)

        v = t.encode_value(l, p)

        self.assertEqual(v, repr_str.encode())

    def test_table_set_get(self):
        pets = Table(prefix='pets', schema={'breed': str, 'owner': str, 'age': int})
        pets.set('henry', {'breed': 'mutt', 'owner': 'stu', 'age': 13})

        henry_all = pets.get('henry')
        henry_single = pets.get('henry', ('age',))
        henry_subset = pets.get('henry', ('owner', 'breed',))

        self.assertDictEqual(henry_all, {'breed': 'mutt', 'owner': 'stu', 'age': 13})
        self.assertDictEqual(henry_single, {'age': 13})
        self.assertDictEqual(henry_subset, {'breed': 'mutt', 'owner': 'stu'})

    def test_table_itemset_itemget(self):
        pets = Table(prefix='pets', schema={'breed': str, 'owner': str, 'age': int})
        pets['black_kitty'] = {'breed': 'fatso', 'owner': 'stu', 'age': 6}

        black_kitty = pets['black_kitty']

        self.assertDictEqual(black_kitty, {'breed': 'fatso', 'owner': 'stu', 'age': 6})

    def test_table_convenience_methods(self):
        p = table(schema={'breed': str, 'age': int})

        self.assertTrue(isinstance(p, TablePlaceholder))
        self.assertDictEqual(p.schema, {'breed': str, 'age': int})

        t = table(prefix='birbs', schema={'breed': str, 'age': int})
        self.assertTrue(isinstance(t, Table))

    def test_table_placeholder_validity(self):
        p = table(schema={'breed': str, 'age': int})
        t = table(prefix='birbs', schema={'breed': str, 'age': int})

        self.assertTrue(p.valid(t))

        bad_t = table(prefix='bribs', schema={'kaw': int, 'age': str})

        self.assertFalse(p.valid(bad_t))

    def test_table_list_placeholder_not_valid(self):
        lp = hlist()

        bad_l = hlist('BAD', value_type=str)

        self.assertFalse(lp.valid(bad_l))

    def test_table_repr_reconstruction(self):
        t = table(prefix='something', schema={'blah': int, 'blerg': str})

    def test_complex_type_repr(self):
        s = '*table({howdy:int,boiii:*map(str,int)})'
        t = parse_complex_type_repr(s)

        self.assertTrue(t.key_type == str)
        ph = t.schema['boiii']
        ph2 = hmap(key_type=str, value_type=int)

        self.assertEqual(ph.key_type, ph2.key_type)
        self.assertEqual(ph.value_type, ph2.value_type)

    def test_table_type_repr_with_prefix(self):
        s = '*table<seneca.contracts.currency:lazytown>({howdy:int,boiii:*map(str,int)})'
        t = parse_complex_type_repr(s)
        self.assertTrue(t.prefix, 'lazytown')

    def test_none_typing(self):
        s = hmap(prefix='hello', value_type=None)
        s.set('yo', 123)
        self.assertEqual(s.get('yo'), 123)

    def test_key_typing(self):
        s = hmap(prefix='hello', key_type=int)
        s.set(1, 123)
        _s = s.get(1)
        self.assertEqual(_s, 123)

    def test_complex_key_typing(self):
        s = hmap(prefix='hello', key_type=hmap())
        s.set(hmap(prefix='uhoh'), 123)

        self.assertEqual(s.get(hmap(prefix='uhoh')), 123)

    def test_none_typing_list(self):
        s = hlist(prefix='hello', value_type=None)
        s.push(123)
        self.assertEqual(s.pop(), 123)

    def test_key_typing_table(self):
        s = table(prefix='hello123', key_type=int, schema={'test1': int, 'test2': str})
        s.set(1, {'test1': 123, 'test2': 'hello'})
        _s = s.get(1)
        self.assertDictEqual(_s, {'test1': 123, 'test2': 'hello'})

    def test_complex_key_typing_table(self):
        s = table(prefix='hello123', key_type=hlist(), schema={'test1': int, 'test2': str})

        s.set(hlist(prefix='uhoh'), {'test1': 123, 'test2': 'hello'})

        _s = s.get(hlist(prefix='uhoh'))

        self.assertDictEqual(_s, {'test1': 123, 'test2': 'hello'})

    def test_table_representation(self):
        s = '*table<seneca.contracts.currency:lazytown>({howdy:int,boiii:*map(str,int)})'
        _s = table(prefix='lazytown', schema={'howdy': int, 'boiii': hmap()})
        self.assertEqual(s, _s.rep())

    def test_table_placeholder_rep(self):
        s = '*table({howdy:int,boiii:*map(str,int)})'
        _s = table(schema={'howdy': int, 'boiii': hmap()})
        self.assertEqual(s, _s.rep())

    def test_robject_raises_error(self):
        r = RObject()
        with self.assertRaises(NotImplementedError):
            r.rep()

    def test_robject_raises_on_none_type(self):
        with self.assertRaises(AssertionError):
            r = RObject(key_type=None)

    def test_complex_types(self):
        self.assertTrue(is_complex_type(hmap()))
        self.assertTrue(is_complex_type(hmap(prefix='ass')))
        self.assertFalse(is_complex_type('literally anything else'))

    def test_vivify_simple_types(self):
        h = hmap(prefix='yoyo')
        self.assertEqual(h['doesntexist'], 0)

        h = hmap(prefix='yoyo', value_type=str)
        self.assertEqual(h['doesntexist'], '')

        h = hmap(prefix='yoyo', value_type=bool)
        self.assertEqual(h['doesntexist'], False)

    def test_vivify_complex_types(self):
        h = hmap(prefix='yoyo', value_type=hmap())
        _h = h['sdgdfgdfg']
        self.assertEqual(_h.value_type, int)
        self.assertEqual(_h.key_type, str)
        self.assertEqual(_h.prefix, 'yoyo.sdgdfgdfg')

        z = h['sdgdfgdfg']['blah']
        self.assertEqual(z, 0)

    def test_vivify_complex_list(self):
        h = hlist(prefix='yoyo', value_type=hlist())
        _h = h[0]
        self.assertEqual(_h.value_type, int)
        self.assertEqual(_h.prefix, 'yoyo.0')

        z = h[0][0]
        self.assertEqual(z, 0)

    def test_ranked_add_get(self):
        r = Ranked(prefix='hello_there')
        r.add('stu', 1000)
        r.add('davis', 100)

        self.assertEqual(r.get_max(), 'stu')
        self.assertEqual(r.get_min(), 'davis')

    def test_ranked_delete(self):
        r = Ranked(prefix='hello_there2')
        r.add('stu', 1000)
        r.add('davis', 100)

        r.delete('stu')

        self.assertEqual(r.get_max(), 'davis')

    def test_ranked_increment(self):
        r = Ranked(prefix='delegates')

        r.increment('stu', 165)

        self.assertEqual(r.score('stu'), 165.0)

        r.delete('stu')

    def test_ranked_rep(self):
        r = Ranked('testing')
        r_str = r.rep()

        _r = parse_representation(r_str)

        self.assertEqual(_r.prefix, r.prefix)
        self.assertEqual(_r.key_type, r.key_type)
        self.assertEqual(_r.value_type, r.value_type)

    def test_drop(self):
        l = HList('something', str)
        l.push('yoyoyo')
        l.push('supsupsup')

        l.drop()

        self.assertEqual(l.pop(), None)

    def test_ranked_pop_max(self):
        r = ranked('woot', str, int)
        r.add('stu', 1000)
        r.add('davis', 1001)
        r.add('falcon', 999)

        _r = r.pop_max()
        self.assertEqual(_r, 'davis')

    def test_ranked_pop_min(self):
        r = ranked('woot', str, int)
        r.add('stu', 1000)
        r.add('davis', 1001)
        r.add('falcon', 999)

        _r = r.pop_min()
        self.assertEqual(_r, 'falcon')

    def test_exists_hlist(self):
        l = HList('bleh', str)
        l.push('stu')
        print(l.exists('stu'))

    def test_float_hmap(self):
        h = hmap('test', str, float)
        h.set('stu', 0.01)

        f = h.get('stu')

        self.assertTrue(isinstance(f, Decimal))
        self.assertEqual(Decimal('0.01'), f)

    def test_float_as_key_type(self):
        h = hmap('test3', float, float)
        h.set(0.1234, 22/7)

        f = h.get(0.1234)

        _f = make_decimal(22/7)

        self.assertTrue(isinstance(f, Decimal))
        self.assertEqual(_f, f)

    def test_simple_type_reprs(self):
        f = parse_representation('float')
        self.assertEqual(f, Decimal)

        f = parse_representation('bytes')
        self.assertEqual(f, bytes)

        f = parse_representation('bool')
        self.assertEqual(f, bool)

    def test_build_placeholder_list_from_repr(self):
        r = '*list(int)'
        l = build_list_from_repr(r)
        self.assertTrue(isinstance(l, ListPlaceholder))
        self.assertEqual(l.value_type, int)
        self.assertEqual(l.rep(), r)

    def test_build_placeholder_ranked_from_repr(self):
        r = '*ranked(int,str)'
        _r = build_ranked_from_repr(r)
        print(_r)
        self.assertTrue(isinstance(_r, RankedPlaceholder))
        self.assertEqual(_r.value_type, str)

if __name__ == '__main__':
    unittest.main()
